/*****************************************************************************
  Copyright (c) 2006 EMC Corporation.
  Copyright (c) 2011 Factor-SPE

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  The full GNU General Public License is included in this distribution in the
  file called LICENSE.

  Authors: Srinivas Aji <Aji_Srinivas@emc.com>
  Authors: Vitalii Demianets <vitas@nppfactor.kiev.ua>

******************************************************************************/

#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/param.h>
//#include <linux/if_bridge.h>
#include <asm/byteorder.h>
#include <utime.h>

#include "bridge_ctl.h"
#include "ctl_functions.h"
#include "netif_utils.h"
#include "packet.h"
#include "log.h"
#include "mstp.h"
#include "driver.h"
#include "libnetlink.h"
#include "clag.h"

#define IFF_LOWER_UP    0x10000     /* driver signals L1 up     */
#define READY_FILE_TIMEOUT 15 /* 15 seconds */
extern clag_role_t clag_role;
extern __u8 clag_system_mac[ETH_ALEN];
LIST_HEAD(bridges);
static int ready_file_ctr = 1;

static bridge_t * create_br(int if_index)
{
    bridge_t *br;
    ifi_t *ifp;

    TST((br = calloc(1, sizeof(*br))) != NULL, NULL);

    /* Init system dependent info */
    br->sysdeps.if_index = if_index;
    if_indextoname(if_index, br->sysdeps.name);
    get_hwaddr(br->sysdeps.name, br->sysdeps.macaddr);

    INFO("Add bridge %s", br->sysdeps.name);
    if(!MSTP_IN_bridge_create(br, br->sysdeps.macaddr))
    {
        free(br);
        return NULL;
    }

    list_add_tail(&br->list, &bridges);
    /* If clag role is master, set the system mac
     */
    if (clag_role == CLAG_ROLE_MASTER)
    {
        clag_set_sys_mac_br(br, clag_system_mac);
    }

    clag_request_dump();
    TST((ifp = calloc(1, sizeof(*ifp))) != NULL, NULL);
    ifp->if_index = if_index;
    ifp->info = (void *)br;
    if (!hash_table_add(ifi_info, &ifp->if_index, sizeof(if_index), ifp)) {
        ERROR("Failed to add %s, ifindex %d to ifi_info\n",
                        br->sysdeps.name, if_index);
        free(ifp);
    }

    return br;
}

bridge_t * find_br(int if_index)
{
    bridge_t *br = NULL;
    ifi_t *ifp = NULL;

    hash_table_find(ifi_info, &if_index, sizeof(if_index), (void *)&ifp);
    if (ifp)
        br = (bridge_t *) ifp->info;

    return br;
}

static port_t * create_if(bridge_t * br, int if_index)
{
    port_t *ifc;
    ifi_t *ifp;
    clag_if_info_t *clag_if;

    TST((ifc = calloc(1, sizeof(*ifc))) != NULL, NULL);

    /* Init system dependent info */
    ifc->sysdeps.if_index = if_index;
    if_indextoname(if_index, ifc->sysdeps.name);
    get_hwaddr(ifc->sysdeps.name, ifc->sysdeps.macaddr);

    int portno;
    if(0 > (portno = get_bridge_portno(ifc->sysdeps.name)))
    {
        ERROR("Couldn't get port number for %s", ifc->sysdeps.name);
        free(ifc);
        return NULL;
    }
    if((0 == portno) || (portno > MAX_PORT_NUMBER))
    {
        ERROR("Port number for %s is invalid (%d)", ifc->sysdeps.name, portno);
        free(ifc);
        return NULL;
    }

    INFO("Add iface %s as port#%d to bridge %s", ifc->sysdeps.name,
         portno, br->sysdeps.name);
    ifc->bridge = br;
    if (is_vlan(ifc->sysdeps.name, &ifc->vlan, NULL)) {
        LOG("%d is a vlan %d", if_index, ifc->vlan);
    }
    if(!MSTP_IN_port_create_and_add_tail(ifc, portno))
    {
        free(ifc);
        return NULL;
    }

    if (!clag_find_isl())
        clag_request_dump();

    if (clag_if_ht) {
        hash_table_find(clag_if_ht, ifc->sysdeps.name, strlen(ifc->sysdeps.name),
                        (void **)&clag_if);
        if (clag_if) {
            clag_set_dual_src_addr(true, clag_if->dual_src_addr,
                                   clag_if->prt_name, clag_if->remote_port_id);
        }
    }

    TST((ifp = calloc(1, sizeof(*ifp))) != NULL, NULL);
    ifp->if_index = if_index;
    ifp->info = (void *)ifc;
    if (!hash_table_add(ifi_info, &ifp->if_index, sizeof(if_index), ifp)) {
        ERROR("Failed to add %s, ifindex %d to ifi_info\n",
                        ifc->sysdeps.name, if_index);
        free(ifp);
    }

    return ifc;
}

port_t * find_if(bridge_t * br, int if_index)
{
    port_t *prt = NULL;
    ifi_t *ifp = NULL;

    hash_table_find(ifi_info, &if_index, sizeof(if_index), (void *)&ifp);
    if (ifp) {
        prt = (port_t *) ifp->info;
        if (br && (br != prt->bridge)) {
            prt = NULL;
        }
    }

    return prt;
}

static inline void delete_if(port_t * ifc)
{
    __u8 zero_mac[ETH_ALEN] = {0x0,0x0,0x0,0x0,0x0,0x0};
    ifi_t *ifp;
    clag_if_info_t *clag_if;

    INFO("Port %s", ifc->sysdeps.name);
    ifc->del = true;
    if (clag_is_isl_port(ifc) && clag_role_slave() &&
            ifc->bridge)
        clag_set_sys_mac_br(ifc->bridge, zero_mac);
    if (clag_if_ht) {
        hash_table_delete(clag_if_ht, ifc->sysdeps.name,
                          strlen(ifc->sysdeps.name), (void **)&clag_if);
        if (clag_if)
            free(clag_if);
    }

    list_del(&ifc->br_list);
    MSTP_IN_delete_port(ifc);
    if (hash_table_delete(ifi_info, &ifc->sysdeps.if_index,
                sizeof(ifc->sysdeps.if_index), (void **)&ifp) == false)
        LOG("Failed to delete ifi %d from ifi_info\n", ifc->sysdeps.if_index);
    else
        free(ifp);

    free(ifc);
}

static inline bool delete_if_byindex(bridge_t * br, int if_index)
{
    port_t *ifc;
    if(!(ifc = find_if(br, if_index)))
        return false;
    delete_if(ifc);
    return true;
}

static bool delete_br_byindex(int if_index)
{
    bridge_t *br;
    ifi_t *ifp;

    if(!(br = find_br(if_index)))
        return false;
    INFO_BRNAME(br, " ");
    list_del(&br->list);
    MSTP_IN_delete_bridge(br);
    if (hash_table_delete(ifi_info, &if_index, sizeof(if_index),
                (void **)&ifp) == false)
        LOG("Failed to delete ifindex %d from ifi_info\n", if_index);
    else
        free(ifp);

    free(br);
    return true;
}

void bridge_one_second(void)
{
    bridge_t *br;

    ready_file_ctr++;
    if (ready_file_ctr == READY_FILE_TIMEOUT) {
        utime(MSTPD_READY_FILE, NULL);
        ready_file_ctr = 1;
    }
    clag_client_one_second();

    list_for_each_entry(br, &bridges, list)
        MSTP_IN_one_second(br);
}

/* New MAC address is stored in addr, which also holds the old value on entry.
   Return true if the address changed */
static bool check_mac_address(char *name, __u8 *addr)
{
    __u8 temp_addr[ETH_ALEN];
    if(get_hwaddr(name, temp_addr))
    {
        LOG("Error getting hw address: %s", name);
        /* Error. Ignore the new value */
        return false;
    }
    if(memcmp(addr, temp_addr, sizeof(temp_addr)) == 0)
        return false;
    else
    {
        memcpy(addr, temp_addr, sizeof(temp_addr));
        return true;
    }
}

static void set_br_up(bridge_t * br, bool up)
{
    bool changed = false;

    if(up != br->sysdeps.up)
    {
        INFO("Set bridge %s %s", br->sysdeps.name, up ? "up" : "down");
        br->sysdeps.up = up;
        changed = true;
        /* When bridge is up, check if vlan filtering is enabled on bridge
         */
        if (up)
            br->sysdeps.vlan_filter = is_bridge_vlan_filter(br->sysdeps.name);
    }

    if(check_mac_address(br->sysdeps.name, br->sysdeps.macaddr))
    {
        LOG("Set bridge %s mac changed", br->sysdeps.name);
        /* MAC address changed */
        /* Notify bridge address change */
        if (!clag_role_configured())
            MSTP_IN_set_bridge_address(br, br->sysdeps.macaddr);
    }

    if(changed)
        MSTP_IN_set_bridge_enable(br, br->sysdeps.up);
}

static void set_if_up(port_t * ifc, bool up, unsigned long *vlan_bitmap,
                    int *pvid)
{
    int speed = -1;
    int duplex = -1;
    bool changed = false;
    char bond_name[IFNAMSIZ];
    int vlan_id = -1;

    memset(bond_name, 0, IFNAMSIZ);
    if(!ifc->del && check_mac_address(ifc->sysdeps.name, ifc->sysdeps.macaddr))
    {
        /* MAC address changed */
        if(check_mac_address(ifc->bridge->sysdeps.name,
           ifc->bridge->sysdeps.macaddr))
        {
            LOG("Set bridge %s %s mac changed", ifc->bridge->sysdeps.name,
                        ifc->sysdeps.name);
            /* Notify bridge address change */
            if (!clag_role_configured())
                MSTP_IN_set_bridge_address(ifc->bridge,
                        ifc->bridge->sysdeps.macaddr);
        }
    }

    if(!up)
    { /* Down */
        if(ifc->sysdeps.up)
        {
            ifc->sysdeps.up = false;
            changed = true;
            INFO("Port %s : %s", ifc->sysdeps.name, (up ? "up" : "down"));
        }
    }
    else
    { /* Up */
        int r = ethtool_get_speed_duplex(ifc->sysdeps.name, &speed, &duplex);
        if((r < 0) || (speed < 0))
            speed = 10;
        if((r < 0) || (duplex < 0))
            duplex = 0; /* Assume half duplex */

        if(speed != ifc->sysdeps.speed)
        {
            ifc->sysdeps.speed = speed;
            changed = true;
        }
        if(duplex != ifc->sysdeps.duplex)
        {
            ifc->sysdeps.duplex = duplex;
            changed = true;
        }
        if(!ifc->sysdeps.up)
        {
            ifc->sysdeps.up = true;
            changed = true;
            INFO("Port %s : %s", ifc->sysdeps.name, (up ? "up" : "down"));
        }
        /* find PVID if vlan filtering is enabled
         */
        if (ifc->bridge->sysdeps.vlan_filter) {
            if (pvid && (ifc->sysdeps.pvid != *pvid)) {
                ifc->sysdeps.pvid = *pvid;
                LOG("%s pvid %d", ifc->sysdeps.name, ifc->sysdeps.pvid);
            }
            if (vlan_bitmap) {
                memcpy(ifc->vlan_bitmap, vlan_bitmap, sizeof(ifc->vlan_bitmap));
            }
        }
        /* propagate CLAG config from bond dev to vlan bond dev
         */
        if (!ifc->bridge->sysdeps.vlan_filter &&
            ifc->vlan &&
            is_vlan(ifc->sysdeps.name, &vlan_id, bond_name) &&
            strncmp(bond_name, ifc->sysdeps.vlan_bond_dev, IFNAMSIZ)) {
            strncpy(ifc->sysdeps.vlan_bond_dev, bond_name, IFNAMSIZ);
            clag_config_to_vlan_bond(bond_name, ifc->sysdeps.name);
        }

    }
    if(changed)
        MSTP_IN_set_port_enable(ifc, ifc->sysdeps.up, ifc->sysdeps.speed,
                                ifc->sysdeps.duplex);
}

/* br_index == if_index means: interface is bridge master */
int bridge_notify(int br_index, int if_index, bool newlink, unsigned flags,
                    unsigned long *vlan_bitmap, int *pvid, bool no_flush)
{
    port_t *ifc;
    bridge_t *br = NULL, *other_br;
    bool up = !!(flags & IFF_UP);
    bool running = up && (flags & IFF_RUNNING);
    bool lower = (flags & IFF_LOWER_UP);
    int br_flags = -1;

    LOG("br_index %d, if_index %d, newlink %d, up %d, running %d, lower %d",
        br_index, if_index, newlink, up, running, lower);

    if((br_index >= 0) && (br_index != if_index))
    {
        if(!(br = find_br(br_index)))
            return -2; /* bridge not in list */
        if (!newlink)
            br_flags = get_flags(br->sysdeps.name);
        if(br_flags >= 0)
            set_br_up(br, !!(br_flags & IFF_UP));
    }

    if(br)
    {
        if(!(ifc = find_if(br, if_index)))
        {
            if(!newlink)
            {
                INFO("Got DELLINK for unknown port %d on "
                     "bridge %d", if_index, br_index);
                return -1;
            }
            /* Check if this interface is slave of another bridge */
            list_for_each_entry(other_br, &bridges, list)
            {
                if(other_br != br)
                    if(delete_if_byindex(other_br, if_index))
                    {
                        INFO("Device %d has come to bridge %d. "
                             "Missed notify for deletion from bridge %d",
                             if_index, br_index, other_br->sysdeps.if_index);
                        break;
                    }
            }
            ifc = create_if(br, if_index);
        }
        if(!ifc)
        {
            ERROR("Couldn't create data for interface %d (master %d)",
                  if_index, br_index);
            return -1;
        }
        if(!newlink)
        {
            delete_if(ifc);
            return 0;
        }
        ifc->no_flush = no_flush;
        /*INFO("port %d: no_flush %d\n", if_index, no_flush); */
        set_if_up(ifc, running, vlan_bitmap, pvid); /* And speed and duplex */
    }
    else
    { /* Interface is not a bridge slave */
        if(!newlink)
        {
            /* DELLINK not from bridge means interface unregistered. */
            /* Cleanup removed bridge or removed bridge slave */
            if(!delete_br_byindex(if_index))
                list_for_each_entry(br, &bridges, list)
                {
                    if(delete_if_byindex(br, if_index))
                        break;
                }
            return 0;
        }
        else
        { /* This may be a new link */
            if(br_index == if_index)
            {
                if(!(br = find_br(br_index)))
                    return -2; /* bridge not in list */
                set_br_up(br, up);
            }
        }
    }
    return 0;
}

struct llc_header
{
    __u8 dest_addr[ETH_ALEN];
    __u8 src_addr[ETH_ALEN];
    __be16 len8023;
    __u8 d_sap;
    __u8 s_sap;
    __u8 llc_ctrl;
} __attribute__((packed));

/* PVST LLC HEADER with Cisco SNAP header
 */
struct pvst_llc_header
{
    __u8 dest_addr[ETH_ALEN];
    __u8 src_addr[ETH_ALEN];
    __be16 len8023;
    __u8 d_sap;
    __u8 s_sap;
    __u8 llc_ctrl;
    __u8 llc_org_code[3];
    __be16 llc_pid;
} __attribute__((packed));

/* PVST LLC 802.1Q tagged header with Cisco SNAP
 */
struct pvst_qtag_llc_header
{
    __u8 dest_addr[ETH_ALEN];
    __u8 src_addr[ETH_ALEN];
    __be16 tag_type;
    __be16 vlan;
    __be16 len8023;
    __u8 d_sap;
    __u8 s_sap;
    __u8 llc_ctrl;
    __u8 llc_org_code[3];
    __be16 llc_pid;
} __attribute__((packed));

/* vlan TLV for PVST
 */
struct pvst_vlan_tlv
{
    __u8 vlan_tlv_code[5];
    __be16 vlan_id;
} __attribute__((packed));

/* vlan TLV for RPVST
 */
struct rpvst_vlan_tlv
{
    __u8 vlan_tlv_code[4];
    __be16 vlan_id;
} __attribute__((packed));

/* LLC_PDU_xxx defines snitched from linux/net/llc_pdu.h */
#define LLC_PDU_LEN_U   3   /* header and 1 control byte */
#define LLC_PDU_TYPE_U  3   /* first two bits */

/* 7.12.3 of 802.1D */
#define LLC_SAP_BSPAN   0x42
#define LLC_SAP_SNAP	0xAA
#define PVST_LLC_PDU_LEN 8
#define PVST_LLC_ORG_CISCO { 0x00, 0x00, 0x0c }
#define RPVST_VLAN_TLV_CISCO { 0x00, 0x00, 0x00, 0x02 }
#define PVST_VLAN_TLV_CISCO { 0x00, 0x00, 0x00, 0x00, 0x02 }
#define PVST_LLC_PID 0x010b
#define PVST_QTYPE 0x8100
static const __u8 bridge_group_address[ETH_ALEN] =
{
    0x01, 0x80, 0xc2, 0x00, 0x00, 0x00
};

/* Cisco mac
 */
static const __u8 pvst_bridge_group_address[ETH_ALEN] =
{
    0x01, 0x00, 0x0c, 0xcc, 0xcc, 0xcd
};

void bridge_bpdu_rcv(int if_index, const unsigned char *data, int len)
{
    port_t *ifc = NULL;
    bridge_t *br = NULL;
    per_tree_port_t *cist = NULL;
    port_t *rx_ifc = NULL;
    port_t *ptr = NULL;
    /* Validate Ethernet and LLC header,
     * maybe we can skip this check thanks to Berkeley filter in packet socket?
     */
    struct llc_header *h;
    struct pvst_llc_header *ph;
    unsigned int l;
    int llc_hdr_len;
    int llc_pdu_len;
    bool pvst = false;
    __u16 pvst_vlan = 0;
    bool pvst_in_cst = false;
    per_tree_port_t *tx_cist = NULL;
    per_tree_port_t *rx_cist = NULL;

    LOG("ifindex %d, len %d", if_index, len);
    TST(len > sizeof(struct llc_header),);
    h = (struct llc_header *)data;
    ifc = find_if(br, if_index);
    if(!ifc)
        return;
    br = ifc->bridge;

    if (ifc->bpduFilterPort)
        return;
    /* check for IEEE/Cisco mac and determine the llc hdr and pdu len
     */
    if (0 == memcmp(h->dest_addr, bridge_group_address, ETH_ALEN)) {
        llc_hdr_len = sizeof(*h);
        llc_pdu_len = LLC_PDU_LEN_U;
    } else if (0 == memcmp(h->dest_addr, pvst_bridge_group_address, ETH_ALEN)) {
        /* Drop pvst bpdu on vlan 0, since there would be a IEEE BPDU recvd
         * Dont drop pvst bpdu in CST mode i.e. when vlan filtering is enabled
         */
        if (ifc->vlan == 0 && (!br->sysdeps.vlan_filter)) {
            return;
        }
        pvst = true;
        ph = (struct pvst_llc_header *)data;
        llc_hdr_len = sizeof(*ph);
        llc_pdu_len = PVST_LLC_PDU_LEN;
        if (br->sysdeps.vlan_filter)
            pvst_in_cst = true;
    } else {
        ERROR("invalid mac");
        return;
    }

    /* sanity checks */
    TST(br == ifc->bridge,);
    TSTDBGLOG(ifc->sysdeps.up,);

    l = __be16_to_cpu(h->len8023);
    if (!pvst) {
        TST(l <= ETH_DATA_LEN && l <= len - ETH_HLEN && l >= LLC_PDU_LEN_U, );
        TST(h->d_sap == LLC_SAP_BSPAN && h->s_sap == LLC_SAP_BSPAN &&
            (h->llc_ctrl & 0x3) == LLC_PDU_TYPE_U,);
    } else {
        /* find originating vlan ID in PVST BPDU */
        pvst_vlan = __be16_to_cpu(*(__be16 *) (data + llc_hdr_len +
                        l - llc_pdu_len - 2));
    }
    /* CLAG RX processing on ISL port
     * If clag_role is configured and clag ISL is up and port BPDU received
     * is ISL, walk through all the ports in the bridge to see if the received
     * smac is same as the configured mac on a dually-connected port.
     * If so, change the prt to be the local dually-connected prt.
     */
    if (clag_role_configured() && (clag_is_isl_port(ifc) && clag_isl_up(br))) {
        list_for_each_entry(ptr, &br->ports, br_list)
        {
            if (clag_is_dual_connected(ptr) &&
                (memcmp(h->src_addr, ptr->clagDualSrcAddr, ETH_ALEN) == 0))
            {
                LOG("clag_dual bpdu recvd on ISL for %s", ptr->sysdeps.name);
                rx_ifc = ptr;
                break;
            }
        }
        if (rx_ifc == NULL)
            rx_ifc = ifc;
    } else {
        rx_ifc = ifc;
    }

    if (pvst_in_cst) {
        rx_cist = GET_CIST_PTP_FROM_PORT(rx_ifc);
        if (rx_cist->state != BR_STATE_FORWARDING) {
            LOG("ignore PVST from %s bcos state %d",
                    rx_ifc->sysdeps.name, rx_cist->state);
            return;
        }
        rx_ifc->num_rx_pvst_tnl_bpdu++;

        /* If vlan filter is enabled in CST mode, flood PVST BPDU on all ports
         */
        list_for_each_entry(ptr, &br->ports, br_list) {
            /* dont send BPDU on incoming port
             */
            if (ptr == rx_ifc)
                continue;
            /* dont send PVST BPDU onto ISL
             */
            if (clag_role_configured() && clag_is_isl_port(ptr) &&
                    clag_isl_up(br))
                continue;
            if (!ptr->sysdeps.up || ptr->bpduFilterPort)
                continue;
            tx_cist = GET_CIST_PTP_FROM_PORT(ptr);
            if (tx_cist->state != BR_STATE_FORWARDING) {
                LOG("dont send PVST from %s to %s bcos state %d",
                    rx_ifc->sysdeps.name, ptr->sysdeps.name, tx_cist->state);
                continue;
            }
            if (!test_bit(pvst_vlan, ptr->vlan_bitmap)) {
                LOG("dont send PVST from %s to %s bcos vlan %d not set",
                    rx_ifc->sysdeps.name, ptr->sysdeps.name, pvst_vlan);
                continue;
            }
            LOG("send PVST from %s to %s, vlan %d",
                    rx_ifc->sysdeps.name, ptr->sysdeps.name, pvst_vlan);
            MSTP_OUT_tx_bpdu(ptr, (bpdu_t *)(data + llc_hdr_len),
                    l - llc_pdu_len, false, h->src_addr, true, pvst_vlan);
        }
    } else {
        MSTP_IN_rx_bpdu(rx_ifc,
                /* Don't include LLC header */
                (bpdu_t *)(data + llc_hdr_len), l - llc_pdu_len);
    }
    /* CLAG RX processing on non-ISL ports
     * if bpdu comes from the root port or a dually connected CLAG port,
     * send the bpdu unmodified onto the CLAG ISL, when clag role is
     * configured. If the BPDU is received on ISL, dont redirect.
     */
    cist = GET_CIST_PTP_FROM_PORT(ifc);
    if (clag_role_configured() && (!clag_is_isl_port(ifc)) &&
        ((cist && (cist->role == roleRoot)) || clag_is_dual_connected(ifc)) &&
            (br->bridgeEnabled)) {
        port_t *isl = clag_find_isl_br(br);
        if (isl && isl->sysdeps.up) {
            if (memcmp(ifc->clagDualSrcAddr, "\0\0\0\0\0\0", 6) == 0) {
                LOG("clag_redirect bpdu %s with pkt smac\n", ifc->sysdeps.name);
                MSTP_OUT_tx_bpdu(isl, (bpdu_t *)(data + llc_hdr_len),
                        l - llc_pdu_len,  true, h->src_addr,
                        pvst_in_cst, pvst_vlan);
            } else {
                LOG("clag_redirect bpdu %s with partner mac", ifc->sysdeps.name);
                MSTP_OUT_tx_bpdu(isl, (bpdu_t *)(data + llc_hdr_len),
                        l - llc_pdu_len,  true, ifc->clagDualSrcAddr,
                        pvst_in_cst, pvst_vlan);
            }
        }
    }
}

static int br_set_state(struct rtnl_handle *rth, unsigned ifindex, __u8 state)
{
    struct
    {
        struct nlmsghdr n;
        struct ifinfomsg ifi;
        char buf[256];
    } req;

    memset(&req, 0, sizeof(req));

    req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifinfomsg));
    req.n.nlmsg_flags = NLM_F_REQUEST | NLM_F_REPLACE;
    req.n.nlmsg_type = RTM_SETLINK;
    req.ifi.ifi_family = AF_BRIDGE;
    req.ifi.ifi_index = ifindex;

    addattr8(&req.n, sizeof(req.buf), IFLA_PROTINFO, state);

    return rtnl_talk(rth, &req.n, 0, 0, NULL, NULL, NULL);
}

static int br_flush_port(char *ifname)
{
    char fname[128];
    snprintf(fname, sizeof(fname), "/sys/class/net/%s/brport/flush", ifname);
    int fd = open(fname, O_WRONLY);
    if (0 > fd)
        LOG("Couldn't open flush file %s for write: %m", fname);
    int write_result = write(fd, "1", 1);
    close(fd);
    TSTDBGLOG(1 == write_result, -1);
    return 0;
}

static int br_set_ageing_time(char *brname, unsigned int ageing_time)
{
    char fname[128], str_time[32];
    snprintf(fname, sizeof(fname), "/sys/class/net/%s/bridge/ageing_time",
             brname);
    int fd = open(fname, O_WRONLY);
    TSTM(0 <= fd, -1, "Couldn't open file %s for write: %m", fname);
    int len = sprintf(str_time, "%u", ageing_time * HZ);
    int write_result = write(fd, str_time, len);
    close(fd);
    TSTDBGLOG(len == write_result, -1);
    return 0;
}

/* External actions for MSTP protocol */

void MSTP_OUT_set_state(per_tree_port_t *ptp, int new_state)
{
    char * state_name;
    port_t *ifc = ptp->port;
    bridge_t *br = ifc->bridge;

    if(ptp->state == new_state) {
        INFO_MSTINAME(br, ifc, ptp, "entering %s role", ROLE_STR(ptp->role));
        return;
    }
    ptp->state = driver_set_new_state(ptp, new_state);

    switch(ptp->state)
    {
        case BR_STATE_LISTENING:
            state_name = "listening";
            break;
        case BR_STATE_LEARNING:
            state_name = "learning";
            break;
        case BR_STATE_FORWARDING:
            state_name = "forwarding";
            ifc->num_trans_fwd++;
            break;
        case BR_STATE_BLOCKING:
            state_name = "blocking";
            ifc->num_trans_blk++;
            break;
        default:
        case BR_STATE_DISABLED:
            state_name = "disabled";
            break;
    }
    INFO_MSTINAME(br, ifc, ptp, "entering %s state(%s)", state_name,
            ROLE_STR(ptp->role));
    if ((ptp->role == roleAlternate) && (ifc->restrictedRole))
        ERROR_MSTINAME(br, ifc, ptp, "role %s due to restricted root role",
                ROLE_STR(ptp->role));

    /* Translate new CIST state to the kernel bridge code */
    if((0 == ptp->MSTID) && (!ifc->del))
    { /* CIST */
        if(0 > br_set_state(&rth_state, ifc->sysdeps.if_index, ptp->state))
            LOG_PRTNAME(br, ifc, "Couldn't set kernel bridge state %s",
                    state_name);
    }
}

/* This function initiates process of flushing
 * all entries for the given port in all FIDs for the
 * given tree.
 * When this process finishes, implementation should signal
 * this by calling MSTP_IN_all_fids_flushed(per_tree_port_t *ptp)
 */
void MSTP_OUT_flush_all_fids(per_tree_port_t * ptp)
{
    port_t *ifc = ptp->port;
    bridge_t *br = ifc->bridge;

    /* Translate CIST flushing to the kernel bridge code */
    if((0 == ptp->MSTID) && (!ifc->del))
    { /* CIST */
        if(0 > br_flush_port(ifc->sysdeps.name))
            LOG_PRTNAME(br, ifc,
                          "Couldn't flush kernel bridge forwarding database");
    }
    /* Completion signal MSTP_IN_all_fids_flushed will be called by driver */
    INFO_MSTINAME(br, ifc, ptp, "Flushing forwarding database");
    driver_flush_all_fids(ptp);
}

void MSTP_OUT_set_ageing_time(port_t *prt, unsigned int ageingTime)
{
    unsigned int actual_ageing_time;
    bridge_t *br = prt->bridge;

    actual_ageing_time = driver_set_ageing_time(prt, ageingTime);
    INFO_PRTNAME(br, prt, "Setting new ageing time to %u", actual_ageing_time);

    /*
     * Translate new ageing time to the kernel bridge code.
     * Kernel bridging code does not support per-port ageing time,
     * so set ageing time for the whole bridge.
     */
    if(0 > br_set_ageing_time(br->sysdeps.name, actual_ageing_time))
        ERROR_BRNAME(br, "Couldn't set new ageing time in kernel bridge");
}

void MSTP_OUT_tx_bpdu(port_t * ifc, bpdu_t * bpdu, int size, bool replay,
                        __u8 *src_addr, bool pvst_in_cst, __u16 vlan_id)
{
    char *bpdu_type;
    bridge_t *br = ifc->bridge;
    bool send_pvst = false;
    int hdr_len;
    void *hdr;
    struct pvst_llc_header ph;
    struct pvst_qtag_llc_header pqh;
    struct llc_header h;
    struct pvst_vlan_tlv vtlv;
    struct rpvst_vlan_tlv rvtlv;
    int tot_len;
    int vtlvlen = 0;
    void *vvtlv = NULL;
    tree_t *cist = NULL;
    unsigned int root_portno;

    cist = GET_CIST_TREE(br);
    /* If clag_role is configured and outgoing port is ISL, dont send BPDU,
     * if clag_isl is up. If clag_isl is down and we are the root bridge,
     * send BPDU only in this case.
     */
    if (!replay && (clag_role_configured()) && clag_is_isl_port(ifc)) {
        root_portno = GET_NUM_FROM_PRIO(cist->rootPortId);
        if (clag_isl_up(br)) {
            LOG("clag_isl dont send BPDU %s", ifc->sysdeps.name);
            return;
        } else if (root_portno != 0) {
            LOG("clag_isl dont send BPDU %s root_port %d ",
                    ifc->sysdeps.name, root_portno);
            return;
        } else {
            LOG("clag_isl send BPDU %s root_port %d ",
                    ifc->sysdeps.name, root_portno);
        }
    }
    /* 1. If clag_role is slave and ISL is up and port is dual connected
     * do not sent BPDU
     * 2. If clag role is slave and ISL is down and port is dual connected,
     * send BPDU
     */
    if ((!replay) && (clag_role_slave()) && (clag_is_dual_connected(ifc)))
    {
        if (clag_isl_up(br))
        {
            LOG("clag_slave dont send BPDU %s", ifc->sysdeps.name);
            return;
        }
#if 0 // This is not needed, since we might not need to sync port-id
        else
        {
             INFO("clag_slave ISL down, send BPDU %s bpdu port_id 0x%x"
                     " nbr port_id 0x%x", ifc->sysdeps.name,
                     bpdu->cistPortID,  ifc->clagRemotePortId);
             bpdu->cistPortID = ifc->clagRemotePortId;
        }
#endif
    }

    switch(bpdu->protocolVersion)
    {
        case protoSTP:
            switch(bpdu->bpduType)
            {
                case bpduTypeConfig:
                    bpdu_type = "STP-Config";
                    break;
                case bpduTypeTCN:
                    bpdu_type = "STP-TCN";
                    ifc->num_tx_tcn++;
                    break;
                default:
                    bpdu_type = "STP-UnknownType";
            }
            break;
        case protoRSTP:
            bpdu_type = "RST";
            break;
        case protoMSTP:
            bpdu_type = "MST";
            break;
        default:
            bpdu_type = "UnknownProto";
    }
    LOG_PRTNAME(br, ifc, "sending %s BPDU%s vlan %d", bpdu_type,
        (bpdu->flags & (1 << offsetTc)) ? ", tcFlag":"", ifc->vlan);

    /* Send PVST BPDU only on vlan device interfaces for STP/RSTP
     */
    if ((ifc->vlan != 0) &&
            ((bpdu->protocolVersion == protoRSTP) ||
            ((bpdu->protocolVersion == protoSTP)))) {
        send_pvst = true;
    }
    /* PVST and RPVST Vlan TLV header is different, account for it
     * to decide the header and PDU Length.
     * for pvst BPDU in CST mode, decide if 802.1Q tag needs to be
     * added based on PVID configured on outgoing port.
     */
    if (send_pvst || (pvst_in_cst && (ifc->sysdeps.pvid == vlan_id))) {
        u_int8_t llcorg[] = PVST_LLC_ORG_CISCO;
        u_int8_t vlantlv[] = PVST_VLAN_TLV_CISCO;
        u_int8_t rvlantlv[] = RPVST_VLAN_TLV_CISCO;

        memcpy(ph.dest_addr, pvst_bridge_group_address, ETH_ALEN);
        if ((replay || pvst_in_cst) && src_addr)
            memcpy(ph.src_addr, src_addr, ETH_ALEN);
        else
            memcpy(ph.src_addr, ifc->sysdeps.macaddr, ETH_ALEN);
        ph.d_sap = ph.s_sap = LLC_SAP_SNAP;
        ph.llc_ctrl = LLC_PDU_TYPE_U;
        memcpy(ph.llc_org_code, llcorg, sizeof(llcorg));
        ph.llc_pid = __cpu_to_be16(PVST_LLC_PID);
        if (bpdu->protocolVersion == protoRSTP)  {
            memcpy(rvtlv.vlan_tlv_code, rvlantlv, sizeof(rvlantlv));
            rvtlv.vlan_id = __cpu_to_be16(ifc->vlan);
            vvtlv = (void *)&rvtlv;
            vtlvlen = sizeof(rvtlv);
        } else {
            memcpy(vtlv.vlan_tlv_code, vlantlv, sizeof(vlantlv));
            vtlv.vlan_id = __cpu_to_be16(ifc->vlan);
            vvtlv = (void *)&vtlv;
            vtlvlen = sizeof(vtlv);
        }
        if (pvst_in_cst)
            vtlvlen = 0;
        ph.len8023 = __cpu_to_be16(size + PVST_LLC_PDU_LEN + vtlvlen);
        hdr_len = sizeof(ph);
        hdr = (void *)&ph;
        tot_len = hdr_len + size + vtlvlen;
    } else if (pvst_in_cst && (ifc->sysdeps.pvid != vlan_id)) {
        u_int8_t llcorg[] = PVST_LLC_ORG_CISCO;

        memcpy(pqh.dest_addr, pvst_bridge_group_address, ETH_ALEN);
        if ((replay || pvst_in_cst) && src_addr)
            memcpy(pqh.src_addr, src_addr, ETH_ALEN);
        else
            memcpy(pqh.src_addr, ifc->sysdeps.macaddr, ETH_ALEN);
        pqh.tag_type = __cpu_to_be16(PVST_QTYPE);
        pqh.vlan = __cpu_to_be16(vlan_id);
        pqh.d_sap = pqh.s_sap = LLC_SAP_SNAP;
        pqh.llc_ctrl = LLC_PDU_TYPE_U;
        memcpy(pqh.llc_org_code, llcorg, sizeof(llcorg));
        pqh.llc_pid = __cpu_to_be16(PVST_LLC_PID);
        pqh.len8023 = __cpu_to_be16(size + PVST_LLC_PDU_LEN);
        hdr_len = sizeof(pqh);
        hdr = (void *)&pqh;
        tot_len = hdr_len + size;

    } else {
        /* IEEE BPDU's
         */
        memcpy(h.dest_addr, bridge_group_address, ETH_ALEN);
        if (replay && src_addr)
            memcpy(h.src_addr, src_addr, ETH_ALEN);
        else
            memcpy(h.src_addr, ifc->sysdeps.macaddr, ETH_ALEN);
        h.len8023 = __cpu_to_be16(size + LLC_PDU_LEN_U);
        h.d_sap = h.s_sap = LLC_SAP_BSPAN;
        h.llc_ctrl = LLC_PDU_TYPE_U;
        hdr_len = sizeof(h);
        hdr = (void *)&h;
        tot_len = hdr_len + size;
    }

    struct iovec iov[3] =
    {
        { .iov_base = hdr, .iov_len = hdr_len },
        { .iov_base = bpdu, .iov_len = size },
        { .iov_base = vvtlv, .iov_len = vtlvlen }
    };
    if (pvst_in_cst) {
        ifc->num_tx_pvst_tnl_bpdu++;
    } else {
        ifc->num_tx_bpdu++;
        if (bpdu->flags & (1 << offsetTc))
            ifc->num_tx_tcn++;
    }
    /* Send the Vlan tlv only for PVST/RPVST BPDU's
     */
    packet_send(ifc->sysdeps.if_index, iov, send_pvst ? 3:2, tot_len);
}

/* User interface commands */

#define CTL_CHECK_BRIDGE                                       \
    bridge_t *br = find_br(br_index);                          \
    if(NULL == br)                                             \
    {                                                          \
        ERROR("Couldn't find bridge with index %d", br_index); \
        return -1;                                             \
    }

#define CTL_CHECK_BRIDGE_PORT                                              \
    CTL_CHECK_BRIDGE;                                                      \
    port_t *prt = find_if(br, port_index);                                 \
    if(NULL == prt)                                                        \
    {                                                                      \
        ERROR_BRNAME(br, "Couldn't find port with index %d", port_index);  \
        return -1;                                                         \
    }

#define CTL_CHECK_BRIDGE_TREE                                      \
    CTL_CHECK_BRIDGE;                                              \
    tree_t *tree;                                                  \
    bool found = false;                                            \
    __be16 MSTID = __cpu_to_be16(mstid);                           \
    list_for_each_entry(tree, &br->trees, bridge_list)             \
        if(tree->MSTID == MSTID)                                   \
        {                                                          \
            found = true;                                          \
            break;                                                 \
        }                                                          \
    if(!found)                                                     \
    {                                                              \
        ERROR_BRNAME(br, "Couldn't find MSTI with ID %hu", mstid); \
        return -1;                                                 \
    }

#define CTL_CHECK_BRIDGE_PERTREEPORT                                     \
    CTL_CHECK_BRIDGE_PORT;                                               \
    per_tree_port_t *ptp;                                                \
    bool found = false;                                                  \
    __be16 MSTID = __cpu_to_be16(mstid);                                 \
    list_for_each_entry(ptp, &prt->trees, port_list)                     \
        if(ptp->MSTID == MSTID)                                          \
        {                                                                \
            found = true;                                                \
            break;                                                       \
        }                                                                \
    if(!found)                                                           \
    {                                                                    \
        ERROR_PRTNAME(br, prt, "Couldn't find MSTI with ID %hu", mstid); \
        return -1;                                                       \
    }

int CTL_get_global_status(GlobalStatus *status)
{
    MSTP_IN_get_global_status(status);
    return 0;
}

int CTL_get_cist_bridge_status(int br_index, CIST_BridgeStatus *status,
                               char *root_port_name)
{
    tree_t *cist;
    per_tree_port_t *ptp;

    CTL_CHECK_BRIDGE;
    MSTP_IN_get_cist_bridge_status(br, status);

    /* find root port name by root_port_id */
    cist = GET_CIST_TREE(br);
    *root_port_name = '\0';
    list_for_each_entry(ptp, &cist->ports, tree_list)
        if(ptp->portId == status->root_port_id)
        {
            strncpy(root_port_name, ptp->port->sysdeps.name, IFNAMSIZ);
            break;
        }
    return 0;
}

int CTL_get_msti_bridge_status(int br_index, __u16 mstid,
                               MSTI_BridgeStatus *status, char *root_port_name)
{
    per_tree_port_t *ptp;

    CTL_CHECK_BRIDGE_TREE;
    MSTP_IN_get_msti_bridge_status(tree, status);

    /* find root port name by root_port_id */
    *root_port_name = '\0';
    list_for_each_entry(ptp, &tree->ports, tree_list)
        if(ptp->portId == status->root_port_id)
        {
            strncpy(root_port_name, ptp->port->sysdeps.name, IFNAMSIZ);
            break;
        }
    return 0;
}

int CTL_set_cist_bridge_config(int br_index, CIST_BridgeConfig *cfg)
{
    CTL_CHECK_BRIDGE;
    return MSTP_IN_set_cist_bridge_config(br, cfg);
}

int CTL_set_msti_bridge_config(int br_index, __u16 mstid, __u16 bridge_priority)
{
    CTL_CHECK_BRIDGE_TREE;
    return MSTP_IN_set_msti_bridge_config(tree, bridge_priority);
}

int CTL_get_cist_port_status(int br_index, int port_index,
                             CIST_PortStatus *status)
{
    CTL_CHECK_BRIDGE_PORT;
    MSTP_IN_get_cist_port_status(prt, status);
    return 0;
}

int CTL_get_msti_port_status(int br_index, int port_index, __u16 mstid,
                             MSTI_PortStatus *status)
{
    CTL_CHECK_BRIDGE_PERTREEPORT;
    MSTP_IN_get_msti_port_status(ptp, status);
    return 0;
}

int CTL_set_cist_port_config(int br_index, int port_index,
                             CIST_PortConfig *cfg)
{
    CTL_CHECK_BRIDGE_PORT;
    return MSTP_IN_set_cist_port_config(prt, cfg);
}

int CTL_set_msti_port_config(int br_index, int port_index, __u16 mstid,
                             MSTI_PortConfig *cfg)
{
    CTL_CHECK_BRIDGE_PERTREEPORT;
    return MSTP_IN_set_msti_port_config(ptp, cfg);
}

int CTL_port_mcheck(int br_index, int port_index)
{
    CTL_CHECK_BRIDGE_PORT;
    return MSTP_IN_port_mcheck(prt);
}

int CTL_set_debug_level(int level)
{
    INFO("level %d", level);
    log_level = level;
    return 0;
}

int CTL_set_batch_cmd(int batch_cmd)
{
    if (batch_cmd_enable == batch_cmd)
        return 0;

    INFO("batch_cmd %d", batch_cmd);
    batch_cmd_enable = batch_cmd;
    if (!batch_cmd_enable)
        MSTP_IN_batch_cmd_stop();

    return 0;
}

int CTL_get_mstilist(int br_index, int *num_mstis, __u16 *mstids)
{
    CTL_CHECK_BRIDGE;
    return MSTP_IN_get_mstilist(br, num_mstis, mstids) ? 0 : -1;
}

int CTL_create_msti(int br_index, __u16 mstid)
{
    CTL_CHECK_BRIDGE;
    if((!driver_create_msti(br, mstid)) || (!MSTP_IN_create_msti(br, mstid)))
        return -1;
    return 0;
}

int CTL_delete_msti(int br_index, __u16 mstid)
{
    CTL_CHECK_BRIDGE;
    if((!driver_delete_msti(br, mstid)) || (!MSTP_IN_delete_msti(br, mstid)))
        return -1;
    return 0;
}

int CTL_get_mstconfid(int br_index, mst_configuration_identifier_t *cfg)
{
    CTL_CHECK_BRIDGE;
    *cfg = br->MstConfigId;
    return 0;
}

int CTL_set_mstconfid(int br_index, __u16 revision, char *name)
{
    CTL_CHECK_BRIDGE;
    MSTP_IN_set_mst_config_id(br, revision, (__u8 *)name);
    return 0;
}

int CTL_get_vids2fids(int br_index, __u16 *vids2fids)
{
    CTL_CHECK_BRIDGE;
    memcpy(vids2fids, br->vid2fid, sizeof(br->vid2fid));
    return 0;
}

int CTL_get_fids2mstids(int br_index, __u16 *fids2mstids)
{
    CTL_CHECK_BRIDGE;
    int i;
    for(i = 0; i < COUNT_OF(br->fid2mstid); ++i)
        fids2mstids[i] = __be16_to_cpu(br->fid2mstid[i]);
    return 0;
}

int CTL_set_vid2fid(int br_index, __u16 vid, __u16 fid)
{
    CTL_CHECK_BRIDGE;
    return MSTP_IN_set_vid2fid(br, vid, fid) ? 0 : -1;
}

int CTL_set_fid2mstid(int br_index, __u16 fid, __u16 mstid)
{
    CTL_CHECK_BRIDGE;
    return MSTP_IN_set_fid2mstid(br, fid, mstid) ? 0 : -1;
}

int CTL_set_vids2fids(int br_index, __u16 *vids2fids)
{
    CTL_CHECK_BRIDGE;
    return MSTP_IN_set_all_vids2fids(br, vids2fids) ? 0 : -1;
}

int CTL_set_fids2mstids(int br_index, __u16 *fids2mstids)
{
    CTL_CHECK_BRIDGE;
    return MSTP_IN_set_all_fids2mstids(br, fids2mstids) ? 0 : -1;
}

int CTL_add_bridges(int *br_array, int* *ifaces_lists)
{
    int i, j, ifcount, brcount = br_array[0];
    bridge_t *br, *other_br;
    port_t *ifc, *nxt;
    int br_flags, if_flags;
    int *if_array;
    bool found;

    for(i = 1; i <= brcount; ++i)
    {
        if(NULL == (br = find_br(br_array[i])))
        {
            if(NULL == (br = create_br(br_array[i])))
            {
                ERROR("Couldn't create data for bridge interface %d",
                      br_array[i]);
                return -1;
            }
            if(0 <= (br_flags = get_flags(br->sysdeps.name)))
                set_br_up(br, !!(br_flags & IFF_UP));
        }
        if_array = ifaces_lists[i - 1];
        ifcount = if_array[0];
        /* delete all interfaces which are not in list */
        list_for_each_entry_safe(ifc, nxt, &br->ports, br_list)
        {
            found = false;
            for(j = 1; j <= ifcount; ++j)
            {
                if(ifc->sysdeps.if_index == if_array[j])
                {
                    found = true;
                    break;
                }
            }
            if(!found)
                delete_if(ifc);
        }
        /* add all new interfaces from the list */
        for(j = 1; j <= ifcount; ++j)
        {
            if(NULL != find_if(br, if_array[j]))
                continue;
            /* Check if this interface is slave of another bridge */
            list_for_each_entry(other_br, &bridges, list)
            {
                if(other_br != br)
                    if(delete_if_byindex(other_br, if_array[j]))
                    {
                        INFO("Device %d has come to bridge %s. "
                             "Missed notify for deletion from bridge %s",
                             if_array[j], br->sysdeps.name,
                             other_br->sysdeps.name);
                        break;
                    }
            }
            if(NULL == (ifc = create_if(br, if_array[j])))
            {
                INFO("Couldn't create data for interface %d (master %s)",
                     if_array[j], br->sysdeps.name);
                continue;
            }
            if(0 <= (if_flags = get_flags(ifc->sysdeps.name)))
                set_if_up(ifc, ((IFF_UP | IFF_RUNNING) ==
                            (if_flags & (IFF_UP | IFF_RUNNING))), NULL, NULL);
        }
    }

    return 0;
}

int CTL_del_bridges(int *br_array)
{
    int i, brcount = br_array[0];

    for(i = 1; i <= brcount; ++i)
        delete_br_byindex(br_array[i]);

    return 0;
}
