/*
 * brmon.c      RTnetlink listener.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 * Authors: Stephen Hemminger <shemminger@osdl.org>
 * Modified by Srinivas Aji <Aji_Srinivas@emc.com>
 *    for use in RSTP daemon. - 2006-09-01
 * Modified by Vitalii Demianets <vitas@nppfactor.kiev.ua>
 *    for use in MSTP daemon. - 2011-07-18
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <netinet/in.h>
//#include <linux/if_bridge.h>

#include "libnetlink.h"
#include "bridge_ctl.h"
#include "netif_utils.h"
#include "epoll_loop.h"
#include "log.h"

/* RFC 2863 operational status */
enum
{
    IF_OPER_UNKNOWN,
    IF_OPER_NOTPRESENT,
    IF_OPER_DOWN,
    IF_OPER_LOWERLAYERDOWN,
    IF_OPER_TESTING,
    IF_OPER_DORMANT,
    IF_OPER_UP,
};


/* link modes */
enum
{
    IF_LINK_MODE_DEFAULT,
    IF_LINK_MODE_DORMANT, /* limit upward transition to dormant */
};
#ifdef DEBUG
static const char *port_states[] =
{
    [BR_STATE_DISABLED] = "disabled",
    [BR_STATE_LISTENING] = "listening",
    [BR_STATE_LEARNING] = "learning",
    [BR_STATE_FORWARDING] = "forwarding",
    [BR_STATE_BLOCKING] = "blocking",
};
#endif

/* Bridge management nested attributes
 * [IFLA_AF_SPEC] = {
 *     [IFLA_BRIDGE_FLAGS]
 *     [IFLA_BRIDGE_MODE]
 *     [IFLA_BRIDGE_VLAN_INFO]
 * }
 */
enum {
	IFLA_BRIDGE_FLAGS,
	IFLA_BRIDGE_MODE,
	IFLA_BRIDGE_VLAN_INFO,
	__IFLA_BRIDGE_MAX,
};
#define IFLA_BRIDGE_MAX (__IFLA_BRIDGE_MAX - 1)

struct bridge_vlan_info {
	__u16 flags;
	__u16 vid;
};

#define BRIDGE_VLAN_INFO_MASTER	(1<<0)	/* Operate on Bridge device as well */
#define BRIDGE_VLAN_INFO_PVID	(1<<1)	/* VLAN is PVID, ingress untagged */
#define BRIDGE_VLAN_INFO_UNTAGGED	(1<<2)	/* VLAN egresses untagged */
#define BRIDGE_VLAN_INFO_RANGE_BEGIN	(1<<3) /* VLAN is start of vlan range */
#define BRIDGE_VLAN_INFO_RANGE_END	(1<<4) /* VLAN is end of vlan range */

static struct rtnl_handle rth;
static struct epoll_event_handler br_handler;

struct rtnl_handle rth_state;

static int dump_msg(const struct sockaddr_nl *who, struct nlmsghdr *n,
                    void *arg)
{
#ifdef DEBUG
    FILE *fp = arg;
    char b1[IFNAMSIZ];
#endif
    struct ifinfomsg *ifi = NLMSG_DATA(n);
    struct rtattr * tb[IFLA_MAX + 1];
    int len = n->nlmsg_len;
    int af_family = ifi->ifi_family;
    bool newlink;
    int br_index;
    unsigned long vlan_bitmap[BR_VLAN_BITMAP_LEN];
    bool vlan_set = false;
    int pvid = 0;
    bool isbridge = false;
    bool no_flush = false;

    if(n->nlmsg_type == NLMSG_DONE)
        return 0;

    len -= NLMSG_LENGTH(sizeof(*ifi));
    if(len < 0)
    {
        ERROR("len- %d", len);
        return -1;
    }

    if(af_family != AF_BRIDGE && af_family != AF_UNSPEC)
        return 0;

    if(n->nlmsg_type != RTM_NEWLINK && n->nlmsg_type != RTM_DELLINK)
        return 0;

    parse_rtattr_flags(tb, IFLA_MAX, IFLA_RTA(ifi), len, NLA_F_NESTED);

    if (tb[IFLA_IFNAME])
        isbridge = is_bridge((char*)RTA_DATA(tb[IFLA_IFNAME]));

    /* Check if we got this from bonding */
    if(tb[IFLA_MASTER] && af_family != AF_BRIDGE && !isbridge)
        return 0;

    if(tb[IFLA_IFNAME] == NULL)
    {
        ERROR("len- %d, if_index- %d", len, ifi->ifi_index);
        return -1;
    }
#ifdef DEBUG
    if(n->nlmsg_type == RTM_DELLINK)
        fprintf(fp, "Deleted ");

    fprintf(fp, "%d: %s ", ifi->ifi_index, (char*)RTA_DATA(tb[IFLA_IFNAME]));

    if(tb[IFLA_OPERSTATE])
    {
        int state = *(int*)RTA_DATA(tb[IFLA_OPERSTATE]);
        switch (state)
        {
            case IF_OPER_UNKNOWN:
                fprintf(fp, "Unknown ");
                break;
            case IF_OPER_NOTPRESENT:
                fprintf(fp, "Not Present ");
                break;
            case IF_OPER_DOWN:
                fprintf(fp, "Down ");
                break;
            case IF_OPER_LOWERLAYERDOWN:
                fprintf(fp, "Lowerlayerdown ");
                break;
            case IF_OPER_TESTING:
                fprintf(fp, "Testing ");
                break;
            case IF_OPER_DORMANT:
                fprintf(fp, "Dormant ");
                break;
            case IF_OPER_UP:
                fprintf(fp, "Up ");
                break;
            default:
                fprintf(fp, "State(%d) ", state);
        }
    }

    if(tb[IFLA_MTU])
        fprintf(fp, "mtu %u ", *(int*)RTA_DATA(tb[IFLA_MTU]));

    if(tb[IFLA_MASTER])
    {
        fprintf(fp, "master %s ",
                if_indextoname(*(int*)RTA_DATA(tb[IFLA_MASTER]), b1));
    }

    if(tb[IFLA_PROTINFO])
    {
        uint8_t state = *(uint8_t *)RTA_DATA(tb[IFLA_PROTINFO]);
        if(state <= BR_STATE_BLOCKING)
            fprintf(fp, "state %s", port_states[state]);
        else
            fprintf(fp, "state (%d)", state);
    }

    fprintf(fp, "\n");
    fflush(fp);
#endif

    if(tb[IFLA_PROTINFO])
    {
        if (tb[IFLA_PROTINFO]->rta_type & NLA_F_NESTED) {
            struct rtattr *prtb[IFLA_BRPORT_MAX+1];

            parse_rtattr_nested(prtb, IFLA_BRPORT_MAX, tb[IFLA_PROTINFO]);

            if (prtb[IFLA_BRPORT_DOWN_PEERLINK_REDIRECT]) {
                no_flush = *(uint8_t *)RTA_DATA(prtb[IFLA_BRPORT_DOWN_PEERLINK_REDIRECT]);
                LOG("no flush: %d\n", no_flush);
            }

            if (prtb[IFLA_BRPORT_LEARNING]) {
                uint8_t learn = *(uint8_t *)RTA_DATA(prtb[IFLA_BRPORT_LEARNING]);
                LOG("learning: %d\n", learn);
            }
        }
    }

    newlink = (n->nlmsg_type == RTM_NEWLINK);

    memset(vlan_bitmap, 0, sizeof(vlan_bitmap));
    if(tb[IFLA_MASTER] && !isbridge) {
        br_index = *(int*)RTA_DATA(tb[IFLA_MASTER]);
        /* If vlan info is present, parse it for bridge ports
         */
        if (tb[IFLA_AF_SPEC]) {
            struct rtattr *i, *list = tb[IFLA_AF_SPEC];
            int rem = RTA_PAYLOAD(list);
            uint16_t vid_range_start = 0;

            for (i = RTA_DATA(list); RTA_OK(i, rem); i = RTA_NEXT(i, rem)) {
                struct bridge_vlan_info *vinfo;
                if (i->rta_type != IFLA_BRIDGE_VLAN_INFO)
                    continue;
                vinfo = RTA_DATA(i);
                if (!vinfo || !vinfo->vid)
                    continue;
                if (vinfo->flags & BRIDGE_VLAN_INFO_RANGE_BEGIN) {
                    vid_range_start = vinfo->vid;
                    continue;
                }
                if (vinfo->flags & BRIDGE_VLAN_INFO_RANGE_END) {}
                else {
                    vid_range_start = vinfo->vid;
                    if (vinfo->flags & BRIDGE_VLAN_INFO_PVID)
                        pvid = vinfo->vid;
                }
                for (; vid_range_start <= vinfo->vid; vid_range_start++) {
                    set_bit(vid_range_start, vlan_bitmap);
                    vlan_set = true;
                }
            }
        }
    } else if(isbridge)
        br_index = ifi->ifi_index;
    else
        br_index = -1;

    bridge_notify(br_index, ifi->ifi_index, newlink, ifi->ifi_flags,
                    vlan_set?vlan_bitmap:NULL, &pvid, no_flush);

    return 0;
}

static inline void br_ev_handler(uint32_t events, struct epoll_event_handler *h)
{
    int ret_val = 0;

    if((ret_val = rtnl_listen(&rth, dump_msg, stdout)) < 0)
    {
        LOG("ret_val = %d", ret_val);
    }
}

int init_bridge_ops(void)
{
    if(rtnl_open(&rth, RTMGRP_LINK) < 0)
    {
        ERROR("Couldn't open rtnl socket for monitoring\n");
        return -1;
    }

    if(rtnl_open(&rth_state, 0) < 0)
    {
        ERROR("Couldn't open rtnl socket for setting state\n");
        return -1;
    }

    if(rtnl_wilddump_request(&rth, PF_BRIDGE, RTM_GETLINK) < 0)
    {
        ERROR("Cannot send dump request: %m\n");
        return -1;
    }

    if(rtnl_dump_filter(&rth, dump_msg, stdout, NULL, NULL) < 0)
    {
        ERROR("Dump terminated\n");
        return -1;
    }

    if(fcntl(rth.fd, F_SETFL, O_NONBLOCK) < 0)
    {
        ERROR("Error setting O_NONBLOCK: %m\n");
        return -1;
    }

    br_handler.fd = rth.fd;
    br_handler.arg = NULL;
    br_handler.handler = br_ev_handler;

    if(add_epoll(&br_handler) < 0)
        return -1;

    return 0;
}
