#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <errno.h>
#include "epoll_loop.h"
#include "mstp.h"
#include "log.h"
#include "clag.h"
#include <asm/byteorder.h>

#define SYSFS_PATH_MAX 256
int clag_role = CLAG_ROLE_NONE;
__u8 clag_system_mac[ETH_ALEN] = {0x0,0x0,0x0,0x0,0x0,0x0};
#define CLAG_CMD_STR_LEN 512
#define FOREACH_PORT_IN_BRIDGE(port, bridge) \
    list_for_each_entry((port), &(bridge)->ports, br_list)
extern struct list_head bridges;
extern void prt_state_machines_begin(port_t *prt);
extern port_t * find_if(bridge_t * br, int if_index);
extern bridge_t * find_br(int if_index);

#define set_socket_address(sa, string) do{  \
    struct sockaddr_un * tmp_sa = (sa);     \
    memset(tmp_sa, 0, sizeof(*tmp_sa));     \
    tmp_sa->sun_family = AF_UNIX;           \
    strcpy(tmp_sa->sun_path, (string));     \
  }while(0)

#define CLAGD_SOCK_NAME         "/var/run/clagd.socket"
#define CLAGD_PIDFILE           "/var/run/clagd.pid"
#define CLAG_CLIENT_TIMEOUT_MAX 10
#define CLAG_CLIENT_BUF_SIZ     512
static bool clag_client_connected = false;
static int clag_client_timeout = 70;
static struct epoll_event_handler clag_handler = {-1};
static bool clag_oper_up = false;
hash_table_t *clag_if_ht = NULL;

bool inline clag_role_configured()
{
    if (clag_role != CLAG_ROLE_NONE)
        return true;

    return false;
}

bool inline clag_role_master()
{
    if (clag_role == CLAG_ROLE_MASTER)
        return true;

    return false;
}

bool inline clag_role_slave()
{
    if (clag_role == CLAG_ROLE_SLAVE)
        return true;

    return false;
}

bool inline clag_is_isl_port(port_t * prt)
{
    if (prt->clagIsl)
        return true;

    return false;
}

port_t *clag_find_isl_br(bridge_t *br)
{
    port_t *prt = NULL;
    bool found = false;

    FOREACH_PORT_IN_BRIDGE(prt, br)
    {
        if (prt->clagIsl) {
            found = true;
            break;
        }
    }

    return found ? prt : NULL;
}

bool clag_isl_up(bridge_t *br)
{
    port_t *prt;
    prt = clag_find_isl_br(br);
    if (prt && prt->clagIslOperUp)
        return true;

    return false;
}

bool inline clag_is_dual_connected(port_t * prt)
{
    __u8 zero_mac[ETH_ALEN] = {0x0,0x0,0x0,0x0,0x0,0x0};

    if (memcmp(prt->clagDualSrcAddr, zero_mac, ETH_ALEN) != 0)
        return true;

    return false;
}

port_t *find_if_name(char *prt_name)
{
    bridge_t *br;
    port_t *ifc;

    list_for_each_entry(br, &bridges, list)
    {
        list_for_each_entry(ifc, &br->ports, br_list)
        {
            if(strncmp(ifc->sysdeps.name, prt_name, IFNAMSIZ) == 0)
                return ifc;
        }
    }
    return NULL;
}

port_t *clag_find_isl()
{
    bridge_t *br;
    port_t *ifc;

    list_for_each_entry(br, &bridges, list)
    {
        list_for_each_entry(ifc, &br->ports, br_list)
        {
            if(clag_is_isl_port(ifc))
                return ifc;
        }
    }
    return NULL;
}

static void clag_set_isl_oper_state()
{
    bridge_t *br;
    port_t *ifc;

    list_for_each_entry(br, &bridges, list)
    {
        list_for_each_entry(ifc, &br->ports, br_list)
        {
            if(clag_is_isl_port(ifc))
                ifc->clagIslOperUp = clag_oper_up;
        }
    }
}

/* Add/del indicated port as ISL
 */
void clag_set_isl(bool add, char *prt_name)
{
    port_t *prt = NULL;
    per_tree_port_t *cist = NULL;
    __u8 valuePri;

    prt = find_if_name(prt_name);
    if (!prt)
        return;
    cist = GET_CIST_PTP_FROM_PORT(prt);

    if (prt->clagIsl != add)
    {
        prt->clagIsl = add;
        if (add) {
            prt->operEdge = true;
            valuePri = (240/16) << 4;
            SET_PRIORITY_IN_IDENTIFIER(valuePri, cist->portId);
            prt->clagIslOperUp = clag_oper_up;
            if ((clag_role == CLAG_ROLE_MASTER) ||
                    (clag_role == CLAG_ROLE_SLAVE && prt->clagIslOperUp))
                clag_set_sys_mac_br(prt->bridge, clag_system_mac);
        } else {
            prt->operEdge = false;
            cist->portId = __constant_cpu_to_be16(0x8000) | prt->port_number;
        }
        LOG_PRTNAME(prt->bridge, prt, "new-%d", prt->clagIsl);
    }
}

/* Set ISL operational state to up/down
 */
void clag_set_isl_state(bool add, char *isl_oper_up)
{
    port_t *prt;
    bridge_t *br;
    __u8 zero_mac[ETH_ALEN] = {0x0,0x0,0x0,0x0,0x0,0x0};

    if (strncmp(isl_oper_up, "up", 2) == 0)
        clag_oper_up = true;
    else if (strncmp(isl_oper_up, "down", 4) == 0)
        clag_oper_up = false;

    prt = clag_find_isl();
    if (!prt)
        return;

    if (prt->clagIslOperUp != clag_oper_up)
    {
        prt->clagIslOperUp = clag_oper_up;
        clag_set_isl_oper_state();
        LOG_PRTNAME(prt->bridge, prt, "new-%d",clag_oper_up);
        /* when ISL goes down on slave node, reset the bridge mac
         */
        if (!prt->clagIslOperUp && (clag_role == CLAG_ROLE_SLAVE)) {
            list_for_each_entry(br, &bridges, list)
                clag_set_sys_mac_br(br, zero_mac);
        } else if ((prt->clagIslOperUp && (clag_role == CLAG_ROLE_SLAVE)) ||
                (clag_role == CLAG_ROLE_MASTER)) {
            list_for_each_entry(br, &bridges, list)
                clag_set_sys_mac_br(br, clag_system_mac);
        }
    }
}

/* Set the partner mac for a dually-connected port along with remote
 * port-id on neighboring switch.
 */
void clag_set_dual_src_addr(bool add, __u8 *dual_src_addr, char *prt_name,
        __u32 remote_port_id)
{
    port_t *prt = NULL;
    clag_if_info_t *clag_if;

    if (!add) {
        remote_port_id = CLAG_SINGLY_CONNECTED_PORT;
        memset(dual_src_addr, 0, ETH_ALEN);
    }

    hash_table_find(clag_if_ht, prt_name, strlen(prt_name), (void **)&clag_if);
    if (!clag_if) {
        clag_if = calloc(1, sizeof(clag_if_info_t));
        strncpy(clag_if->prt_name, prt_name, IFNAMSIZ);
        hash_table_add(clag_if_ht, clag_if->prt_name, strlen(prt_name),
                       clag_if);
    }
    memcpy(clag_if->dual_src_addr, dual_src_addr, ETH_ALEN);
    clag_if->remote_port_id = remote_port_id;

    prt = find_if_name(prt_name);
    if (!prt)
        return;

    if (memcmp(prt->clagDualSrcAddr, dual_src_addr, ETH_ALEN) != 0) {
        memcpy(prt->clagDualSrcAddr, dual_src_addr, ETH_ALEN);
        LOG_PRTNAME(prt->bridge, prt, "clagDualSrcAddr new-"
                "0x%x:0x%x:0x%x:0x%x:0x%x:0x%x",
                prt->clagDualSrcAddr[0], prt->clagDualSrcAddr[1],
                prt->clagDualSrcAddr[2], prt->clagDualSrcAddr[3],
                prt->clagDualSrcAddr[4], prt->clagDualSrcAddr[5]);
    }

    if (prt->clagRemotePortId != remote_port_id) {
        prt->clagRemotePortId = remote_port_id;
        LOG_PRTNAME(prt->bridge, prt, "clagRemotePortId new-%u",
                    prt->clagRemotePortId);
    }
}

void clag_set_sys_mac_br(bridge_t *br, __u8 *sys_mac)
{
    __u8 zero_mac[ETH_ALEN] = {0x0,0x0,0x0,0x0,0x0,0x0};

    INFO("set %s mac %02x:%02x:%02x:%02x:%02x:%02x\n", br->sysdeps.name,
                sys_mac[0], sys_mac[1], sys_mac[2], sys_mac[3],
                sys_mac[4], sys_mac[5]);

    if (0 == memcmp(zero_mac, sys_mac, ETH_ALEN))
        MSTP_IN_set_bridge_address(br, br->sysdeps.macaddr);
    else
        MSTP_IN_set_bridge_address(br, sys_mac);
}

void clag_set_role(bool add, char *role)
{
    int new_clag_role;
    int old_clag_role = clag_role;
    __u8 zero_mac[ETH_ALEN] = {0x0,0x0,0x0,0x0,0x0,0x0};
    bridge_t *br;
    port_t *isl_port;

    if (!strcmp(role, "primary")) {
        new_clag_role = CLAG_ROLE_MASTER;
    } else if (!strcmp(role, "secondary")) {
        new_clag_role = CLAG_ROLE_SLAVE;
    } else {
        new_clag_role = CLAG_ROLE_NONE;
    }

    if (clag_role != new_clag_role) {
        LOG("clag_role %d, old_clag_role %d", new_clag_role, old_clag_role);
        clag_role = new_clag_role;
        /* none role, reset the sys_mac to zero
         */
        if (clag_role == CLAG_ROLE_NONE) {
            list_for_each_entry(br, &bridges, list)
                clag_set_sys_mac_br(br, zero_mac);
        } else {
            /* for slave role when ISL is down or when ISL is not set,
             * ensure that the bridge_id  is reset to bridge driver's
             * bridge_id
             */
            isl_port = clag_find_isl();
            if ((clag_role == CLAG_ROLE_SLAVE) && ((!isl_port) ||
                        (isl_port && !isl_port->clagIslOperUp))) {
                list_for_each_entry(br, &bridges, list)
                    clag_set_sys_mac_br(br, zero_mac);
                return;
            }

            /* Set sys_mac in all other cases
             */
            list_for_each_entry(br, &bridges, list)
                clag_set_sys_mac_br(br, clag_system_mac);
        }
    }
}

void clag_set_system_mac(bool add, __u8 sys_mac[ETH_ALEN])
{
    bridge_t *br;
    port_t *isl_port;

    LOG("set_sys_mac 0x%x:0x%x:0x%x:0x%x:0x%x:0x%x", sys_mac[0], sys_mac[1],
            sys_mac[2], sys_mac[3], sys_mac[4], sys_mac[5]);

    if (memcmp(sys_mac, clag_system_mac, ETH_ALEN) != 0) {
        memcpy(clag_system_mac, sys_mac, ETH_ALEN);
        if (clag_role != CLAG_ROLE_NONE) {
            /* for slave role when ISL is down, dont configure sys_mac
             */
            isl_port = clag_find_isl();
            if ((clag_role == CLAG_ROLE_SLAVE) && ((!isl_port) ||
                (isl_port && !isl_port->clagIslOperUp)))
                return;
            list_for_each_entry(br, &bridges, list)
                clag_set_sys_mac_br(br, clag_system_mac);
        }
    }
}

bool is_clagd_alive(void)
{
    struct stat sts;
    if (stat(CLAGD_PIDFILE, &sts) != 0)
        return false;

    FILE *fp = fopen(CLAGD_PIDFILE, "r");
    if (!fp)
        return false;

    int clagd_pid;
    fscanf(fp, "%d", &clagd_pid);
    fclose(fp);

    char path[64];
    sprintf(path, "/proc/%d", clagd_pid);
    if (stat(path, &sts) == -1 && errno == ENOENT)
        return false;

    LOG("clagd alive!\n");
    return true;
}

/* Cleanup clag configuration when socket connection closes to clagd
 *
 */
void clag_config_cleanup()
{
    bridge_t *br;
    port_t *prt = NULL;
    __u8 zero_mac[ETH_ALEN] = {0x0,0x0,0x0,0x0,0x0,0x0};
    per_tree_port_t *cist = NULL;

    clag_role = CLAG_ROLE_NONE;
    memset(clag_system_mac, 0, ETH_ALEN);
    list_for_each_entry(br, &bridges, list)
    {
        clag_set_sys_mac_br(br, zero_mac);
        FOREACH_PORT_IN_BRIDGE(prt, br)
        {
            if (prt->clagIsl) {
                prt->clagIsl = false;
                cist = GET_CIST_PTP_FROM_PORT(prt);
                cist->portId = __constant_cpu_to_be16(0x8000)|prt->port_number;
            }
            prt->clagIslOperUp = false;
            prt->clagRemotePortId = CLAG_SINGLY_CONNECTED_PORT;
            memset(prt->clagDualSrcAddr, 0, ETH_ALEN);
        }
    }
}

void clag_client_cleanup(void)
{
    if(clag_handler.fd >= 0)
    {
        close(clag_handler.fd);
        clag_handler.fd = -1;
    }
    clag_client_connected = false;
    clag_client_timeout = CLAG_CLIENT_TIMEOUT_MAX;
    if (clag_if_ht) {
        hash_table_free(clag_if_ht, free);
        clag_if_ht = NULL;
    }
    clag_config_cleanup();
    LOG(" ");
}

/* for old bridge driver, propagate the CLAG configuration
 * for physical bond devices to bond vlan devices in other
 * bridges.
 */
void clag_config_to_vlan_bond(char *bond_dev, char *vlan_dev)
{
    port_t *bond = NULL;
    bridge_t *br;
    port_t *ifc;

    if (!bond_dev)
        return;
    bond = find_if_name(bond_dev);
    if (!bond)
        return;

    list_for_each_entry(br, &bridges, list) {
        /* Ignore the bridge, bond is part of
         */
        if (bond->bridge == br)
            continue;
        list_for_each_entry(ifc, &br->ports, br_list) {
            /* if vlan_dev is specified, ignore other vlan_dev */
            if (vlan_dev &&
                    (strncmp(ifc->sysdeps.name, vlan_dev, IFNAMSIZ) != 0))
                continue;
            /* Verify vlan_dev and bond_dev are associated */
            if (strncmp(ifc->sysdeps.vlan_bond_dev, bond_dev, IFNAMSIZ) == 0) {
                if (bond->clagIsl != ifc->clagIsl)
                    clag_set_isl(bond->clagIsl, ifc->sysdeps.name);
                if (memcmp(ifc->clagDualSrcAddr, bond->clagDualSrcAddr,
                            ETH_ALEN) != 0) {
                    clag_set_dual_src_addr(1, bond->clagDualSrcAddr,
                            ifc->sysdeps.name, bond->clagRemotePortId);
                }
            }
        }
    }
}

static void one_clag_cmd(char *cmd)
{
    char *oper, *obj, *v0, *v1, *p;
    bool add = false;
    __u8 mac[ETH_ALEN] = {0x0,0x0,0x0,0x0,0x0,0x0};
    __u32 port_id;
    int pri, pd;

    if (*cmd == '\n')
        return;

    p = cmd;
    while (*p == ' ')
        p++;

    oper = strtok(p, " ");
    if (!oper)
        goto err;

    if (!strcmp(oper, "add"))
        add = true;

    obj = strtok(NULL, " = ");
    if (!obj)
        goto err;

    while (*obj == ' ')
        obj++;

    v0 = obj + strlen(obj) + 3;
    while (*v0 == ' ')
        v0++;

    LOG("obj:<%s>\n", obj);

    if (!strcmp(obj, "peer-link")) {
        /* set peer link */
        INFO("setting (%d) peer link: %s\n", add, v0);
        clag_set_isl(add, v0);
        clag_config_to_vlan_bond(v0, NULL);
    } else if (!strcmp(obj, "clag-state")) {
        /* set clag state */
        INFO("setting (%d) clag state: %s\n", add, v0);
        clag_set_isl_state(add, v0);
    } else if (!strncmp(obj, "mac-", 4)) {
        v1 = strchr(v0, ',');
        if (v1) {
            *v1 = '\0';
            v1++;
            while (*v1 == ' ')
                v1++;
        } else
            goto err;

        /* set the mac */

        INFO("setting (%d) mac %s <%s, %s>\n", add, obj+4, v0, v1);
        sscanf(obj+4, "%2hhx:%2hhx:%2hhx:%2hhx:%2hhx:%2hhx",
            &mac[0], &mac[1], &mac[2], &mac[3], &mac[4], &mac[5]);
        if (!strcmp(v1, "None"))
            port_id = CLAG_SINGLY_CONNECTED_PORT;
        else {
            sscanf(v1, "%d.%d", &pri, &pd);
            port_id = (pri << 12) | pd;
            LOG("clag pri 0x%x port_id 0x%x\n", pri, port_id);
        }
        clag_set_dual_src_addr(add, mac, v0, port_id);
        clag_config_to_vlan_bond(v0, NULL);
    } else if (!strcmp(obj, "system-mac")) {
        INFO("setting system-mac %s\n", v0);
        sscanf(v0, "%2hhx:%2hhx:%2hhx:%2hhx:%2hhx:%2hhx",
            &mac[0], &mac[1], &mac[2], &mac[3], &mac[4], &mac[5]);
        clag_set_system_mac(add, mac);
    } else if (!strcmp(obj, "clag-role")) {
        INFO("setting clag-role %s\n", v0);
        clag_set_role(add, v0);
    } else
        goto err;

    return;

err:
    ERROR("malformed command %s\n", cmd);
}

static void get_clag_cmd(char *input, char *newcmd)
{
    char *ptr;
    int i;

    ptr = input;
    i = 0;
    while (*ptr != '\n') {
        newcmd[i] = *ptr;
        ptr++;
        i++;
    }
    newcmd[i] = '\0';
}

static void process_clag_cmds(char *input, int len)
{
    char *cmd, *ptr;
    char newcmd[1024];

    if (!input)
        return;

    //LOG("input %s\n", input);
    cmd = input;
    get_clag_cmd(input, newcmd);

    while (cmd) {
        LOG("newcmd %s\n", newcmd);
        ptr = cmd + strlen(newcmd) + 1;
        if (strlen(newcmd) < 4)
            break;
        one_clag_cmd(newcmd);
        if ((ptr - input + 1) >= len)
            break;
        cmd = ptr;
        get_clag_cmd(ptr, newcmd);
    }
}

static void clag_ev_handler(uint32_t events, struct epoll_event_handler *h)
{

    LOG("0x%x\n", events);
    if (events & (EPOLLRDHUP | EPOLLERR)) {
        /* socket to clagd is broken, cleanup, will try again */
        clag_client_cleanup();

        /* invoke an operational state down */

        return;
    }

    if (events & EPOLLIN) {
        char *cmd = calloc(1, CLAG_CLIENT_BUF_SIZ), *ptr;
        int rem_size = CLAG_CLIENT_BUF_SIZ;
        int total_size = 0;

        /* read all data */
        ptr = cmd;
        while (1) {
            int count;

            count = read(h->fd, ptr, rem_size - 1);
            if (count == -1) {
                if (errno != EAGAIN) {
                    LOG("read error\n");
                    clag_client_cleanup();
                    if (cmd) free(cmd);
                    return;
                }
                break;
            } else if (count == 0) {
                break;
            } else {
                rem_size -= count;
                total_size += count;
                if (rem_size == 1) {
                    cmd = realloc(cmd, total_size + CLAG_CLIENT_BUF_SIZ + 1);
                    cmd[total_size + CLAG_CLIENT_BUF_SIZ] = '\0';
                    ptr = cmd + total_size;
                    rem_size = CLAG_CLIENT_BUF_SIZ;
                } else
                    ptr += count;
            }
        }
        /* parse and do something */
        process_clag_cmds(cmd, total_size);
        free(cmd);
    }
}

int clag_request_dump(void)
{
    if (clag_handler.fd == -1)
        return 0;

    /* initiatiate the dump request */
    char cmd[] = "dumpmstpstate\0";
    LOG("getall length %d\n", strlen(cmd)+1);
    if (write(clag_handler.fd, cmd, strlen(cmd)+1) == -1) {
        LOG("Write error\n");
        clag_client_cleanup();
        return -1;
    }

    return 0;
}

int clag_client_init(void)
{
    struct sockaddr_un sa_svr;
    int s;

    /* check for clagd pidfile */
    if (!is_clagd_alive()) {
        return -1;
    }

    TST(strlen(CLAGD_SOCK_NAME) < sizeof(sa_svr.sun_path), -1);
    if(0 > (s = socket(PF_UNIX, SOCK_STREAM, 0)))
    {
        LOG("Couldn't open unix socket: %m");
        return -1;
    }

    int on = 1;
    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) < 0) {
        LOG("Couldn't setsockopt: %m");
        close(s);
        return -1;
    }

    set_socket_address(&sa_svr, CLAGD_SOCK_NAME);

    if(0 != connect(s, (struct sockaddr *)&sa_svr, sizeof(sa_svr)))
    {
        LOG("Couldn't connect to server: %m");
        close(s);
        return -1;
    }

    if (fcntl(s, F_SETFL, O_NONBLOCK) < 0)
    {
        LOG("Error setting O_NONBLOCK on clagd_fd\n");
        close(s);
        return -1;
    }

    clag_handler.fd = s;
    clag_handler.arg = NULL;
    clag_handler.handler = clag_ev_handler;

    if (add_epoll(&clag_handler) < 0) {
        LOG("Error adding to epoll\n");
        clag_client_cleanup();
        return -1;
    }

    clag_client_connected = true;
    clag_client_timeout = CLAG_CLIENT_TIMEOUT_MAX;
    LOG("connected to clagd\n");

    clag_if_ht = hash_table_alloc(INTF_HASH_SIZE);

    /* initiatiate the dump request */
    char cmd[] = "getall\0";
    LOG("getall length %d\n", strlen(cmd)+1);
    if (write(clag_handler.fd, cmd, strlen(cmd)+1) == -1) {
        LOG("Write error\n");
        clag_client_cleanup();
        return -1;
    }

    return 0;
}

void clag_client_one_second(void)
{
    if (clag_client_connected)
        return;

    /* During initial start of mstpd try to connect to CLAG every 1s for
     * upto 60seconds
     */
    if (clag_client_timeout > CLAG_CLIENT_TIMEOUT_MAX) {
        clag_client_init();
        --clag_client_timeout;
        return;
    }

    if (--clag_client_timeout == 0) {
        clag_client_timeout = CLAG_CLIENT_TIMEOUT_MAX;
        clag_client_init();
    }
}
