/*****************************************************************************
  Copyright (c) 2011 Factor-SPE

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  The full GNU General Public License is included in this distribution in the
  file called LICENSE.

  Authors: Vitalii Demianets <vitas@nppfactor.kiev.ua>

******************************************************************************/

#include <string.h>
#include <getopt.h>
#include <dirent.h>
#include <errno.h>
#include <limits.h>
#include <sys/stat.h>

#include "json.h"
#include "ctl_socket_client.h"
#include "log.h"

#ifdef  __LIBC_HAS_VERSIONSORT__
#define sorting_func    versionsort
#else
#define sorting_func    alphasort
#endif

static int get_index_die(const char *ifname, const char *doc, bool die, u_char use_json)
{
    int r = if_nametoindex(ifname);
    if(0 == r)
    {
        if (use_json)
          {
            json_object *json_err = json_object_new_object();
            json_object *json_sub = json_object_new_object();
            json_object_string_add(json_sub, "doc", doc);
            json_object_string_add(json_sub, "ifname", ifname);
            json_object_string_add(json_err, "error", "Can't find the index for the doc and interface | Not a valid interface");
            json_object_object_add(json_err, "extraData", json_sub);
            printf("\n%s\n", json_object_to_json_string(json_err));
            json_object_free(json_err);
          }
        else
          fprintf(stderr, "Can't find index for %s %s. Not a valid interface.\n",
                  doc, ifname);
        if(die)
            exit(1);
        return -1;
    }
    return r;
}

static inline int get_index(const char *ifname, const char *doc)
{
    return get_index_die(ifname, doc, true, 0);
}

static inline int get_id(const char *str, const char *doc, unsigned int max_id)
{
    int id = strtol(str, NULL, 10);
    if((0 > id) || (max_id < id)
       || ((0 == id) && ('0' != str[0]))
      )
    {
        fprintf(stderr, "Bad %s %s\n", doc, str);
        return -1;
    }
    return id;
}

#define BOOL_STR(x) ((x) ? "yes" : "no")
#define PROTO_VERS_STR(x)   ((protoRSTP == (x)) ? "rstp" : \
                             ((protoMSTP <= (x)) ? "mstp" : "stp"))

#define CLAG_ROLE_STR(x)   ((1 == (x)) ? "primary" : \
                             ((2 == (x)) ? "secondary" : "unknown"))
typedef enum {
    PARAM_NULL = 0,
    /* bridge params */
    PARAM_ENABLED,
    PARAM_BRID,
    PARAM_DSGNROOT,
    PARAM_REGNROOT,
    PARAM_ROOTPORT,
    PARAM_PATHCOST,
    PARAM_INTPATHCOST,
    PARAM_MAXAGE,
    PARAM_BRMAXAGE,
    PARAM_FWDDELAY,
    PARAM_BRFWDDELAY,
    PARAM_TXHOLDCNT,
    PARAM_MAXHOPS,
    PARAM_FORCEPROTVERS,
    PARAM_TOPCHNGTIME,
    PARAM_TOPCHNGCNT,
    PARAM_TOPCHNGSTATE,
    /* port params */
    PARAM_ROLE,
    PARAM_STATE,
    PARAM_PORTID,
    PARAM_EXTPORTCOST,
    PARAM_ADMINEXTCOST,
    PARAM_INTPORTCOST,
    PARAM_ADMININTCOST,
    PARAM_DSGNEXTCOST,
    PARAM_DSGNRROOT,
    PARAM_DSGNINTCOST,
    PARAM_DSGNBR,
    PARAM_DSGNPORT,
    PARAM_ADMINEDGEPORT,
    PARAM_AUTOEDGEPORT,
    PARAM_OPEREDGEPORT,
    PARAM_TOPCHNGACK,
    PARAM_P2P,
    PARAM_ADMINP2P,
    PARAM_RESTRROLE,
    PARAM_RESTRTCN,
    PARAM_PORTHELLOTIME,
    PARAM_DISPUTED,
    PARAM_BPDUGUARDPORT,
    PARAM_BPDUGUARDERROR,
    PARAM_BRHELLO,
    PARAM_BRAGEING,
    PARAM_NETWORKPORT,
    PARAM_BA_INCONSISTENT,
    PARAM_BPDUFILTERPORT,
} param_id_t;

typedef struct {
    param_id_t id;
    char *str;
} cmd_param_t;

static const cmd_param_t cist_bridge_params[] = {
    { PARAM_ENABLED,      "enabled" },
    { PARAM_BRID,         "bridge-id" },
    { PARAM_DSGNROOT,     "designated-root" },
    { PARAM_REGNROOT,     "regional-root" },
    { PARAM_ROOTPORT,     "root-port" },
    { PARAM_PATHCOST,     "path-cost" },
    { PARAM_INTPATHCOST,  "internal-path-cost" },
    { PARAM_MAXAGE,       "max-age" },
    { PARAM_BRMAXAGE,     "bridge-max-age" },
    { PARAM_FWDDELAY,     "forward-delay" },
    { PARAM_BRFWDDELAY,   "bridge-forward-delay" },
    { PARAM_TXHOLDCNT,    "tx-hold-count" },
    { PARAM_MAXHOPS,      "max-hops" },
    { PARAM_FORCEPROTVERS,"force-protocol-version" },
    { PARAM_TOPCHNGTIME,  "time-since-topology-change" },
    { PARAM_TOPCHNGCNT,   "topology-change-count" },
    { PARAM_TOPCHNGSTATE, "topology-change" },
    { PARAM_BRHELLO,      "hello-time" },
    { PARAM_BRAGEING,     "ageing-time" },
};

static int all = 0;
#define SYSFS_PATH_MAX 256
#define SYSFS_CLASS_NET "/sys/class/net"
static bool is_stp_userspace_bridge(const char *br_name)
{
    char path[SYSFS_PATH_MAX];
    int save_errno;
    bool result;
    struct stat st;
    FILE *fd;
    unsigned char stp_state;

    snprintf(path, SYSFS_PATH_MAX, SYSFS_CLASS_NET "/%s/bridge",
             br_name);
    save_errno = errno;
    result = (0 == stat(path, &st)) && S_ISDIR(st.st_mode);
    errno = save_errno;
    if (result) {
        strcat(path, "/stp_state");
        if ((0 == stat(path, &st)) && (fd = fopen(path, "r"))) {
            fread(&stp_state, 1, 1, fd);
            fclose(fd);
            if (((unsigned int)stp_state - '0') != 2) {
                return 0;
            }
        } else
            result = 0;
    }

    return result;
}

static int do_showbridge(const char *br_name, param_id_t param_id, u_char use_json, json_object *json)
{
    CIST_BridgeStatus s;
    char root_port_name[IFNAMSIZ];
    unsigned int root_portno;
    int br_index = get_index_die(br_name, "bridge", false, use_json);
    json_object *json_bridge = NULL;

    if (use_json)
      json_bridge = json_object_new_object();

    if(0 > br_index)
        return br_index;
    if (is_stp_userspace_bridge(br_name) == 0)
        return 0;
    if(CTL_get_cist_bridge_status(br_index, &s, root_port_name))
        return -1;

    switch(param_id)
    {
        case PARAM_NULL:
            if (use_json)
              {
                char tempA[BUFSIZ];
                if (s.enabled)
                  json_object_boolean_true_add(json_bridge, "enabled");

                sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.bridge_id));
                json_object_string_add(json_bridge, "bridgeId", tempA);

                memset(&tempA[0], 0, sizeof(tempA));
                sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.designated_root));
                json_object_string_add(json_bridge, "dsgnRoot", tempA);

                memset(&tempA[0], 0, sizeof(tempA));
                sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.regional_root));
                json_object_string_add(json_bridge, "regionalRoot", tempA);

                if (0 != (root_portno = GET_NUM_FROM_PRIO(s.root_port_id)))
                  {
                    json_object_string_add(json_bridge, "rootPortName", root_port_name);
                    json_object_int_add(json_bridge, "rootPortNum", root_portno);
                  }

                json_object_int_add(json_bridge, "pathCost", s.root_path_cost);
                json_object_int_add(json_bridge, "intPathCost", s.internal_path_cost);
                json_object_int_add(json_bridge, "maxAge", s.root_max_age);
                json_object_int_add(json_bridge, "bridgeMaxAge", s.bridge_max_age);
                json_object_int_add(json_bridge, "fwdDelay", s.root_forward_delay);
                json_object_int_add(json_bridge, "bridgeFwdDelay", s.bridge_forward_delay);
                json_object_int_add(json_bridge, "txHoldCounter", s.tx_hold_count);
                json_object_int_add(json_bridge, "maxHops", s.max_hops);
                json_object_int_add(json_bridge, "helloTime", s.bridge_hello_time);
                json_object_int_add(json_bridge, "ageingTime", s.Ageing_Time);
                json_object_string_add(json_bridge, "forceProtocolVersion", PROTO_VERS_STR(s.protocol_version));
                json_object_int_add(json_bridge, "timeSinceLastTopoChng", s.time_since_topology_change);
                json_object_int_add(json_bridge, "topoChngCounter", s.topology_change_count);
                if (s.topology_change)
                  json_object_boolean_true_add(json_bridge, "topoChng");
                json_object_string_add(json_bridge, "topoChngPort", s.topology_change_port);
                json_object_string_add(json_bridge, "lastTopoChngPort", s.last_topology_change_port);

                if (all)
                  {
                    json_object_int_add(json_bridge, "migrateTime", s.Migrate_Time);
                    json_object_int_add(json_bridge, "uptime", s.uptime);
                    json_object_string_add(json_bridge, "prssmState", PRSSM_STATE(s.PRSSM_state));
                    json_object_int_add(json_bridge, "ifIndex", s.if_index);
                    json_object_string_add(json_bridge, "name", s.name);
                    if (s.up)
                      json_object_boolean_true_add(json_bridge, "up");
                    if (s.vlan_filter)
                      json_object_boolean_true_add(json_bridge, "vlanFilter");
                  }
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              {
                printf("%s CIST info\n", br_name);
                printf("  enabled         %s\n", BOOL_STR(s.enabled));
                printf("  bridge id       "BR_ID_FMT"\n",
                       BR_ID_ARGS(s.bridge_id));
                printf("  designated root "BR_ID_FMT"\n",
                       BR_ID_ARGS(s.designated_root));
                printf("  regional root   "BR_ID_FMT"\n",
                       BR_ID_ARGS(s.regional_root));
                printf("  root port       ");
                if (0 != (root_portno = GET_NUM_FROM_PRIO(s.root_port_id)))
                  printf("%s (#%u)\n", root_port_name, root_portno);
                else
                  printf("none\n");
                printf("  path cost     %-10u ", s.root_path_cost);
                printf("internal path cost   %u\n", s.internal_path_cost);
                printf("  max age       %-10hhu ", s.root_max_age);
                printf("bridge max age       %hhu\n", s.bridge_max_age);
                printf("  forward delay %-10hhu ", s.root_forward_delay);
                printf("bridge forward delay %hhu\n", s.bridge_forward_delay);
                printf("  tx hold count %-10u ", s.tx_hold_count);
                printf("max hops             %hhu\n", s.max_hops);
                printf("  hello time    %-10u ", s.bridge_hello_time);
                printf("ageing time          %u\n", s.Ageing_Time);
                printf("  force protocol version     %s\n",
                       PROTO_VERS_STR(s.protocol_version));
                printf("  time since topology change %us\n",
                       s.time_since_topology_change);
                printf("  topology change count      %u\n",
                       s.topology_change_count);
                printf("  topology change            %s\n",
                       BOOL_STR(s.topology_change));
                printf("  topology change port       %s\n",
                       s.topology_change_port);
                printf("  last topology change port  %s\n",
                       s.last_topology_change_port);
                if (all)
                  {
                    printf("  migrate_time: %u," \
                           " uptime: %u, PRSSM_state: %s\n",
                           s.Migrate_Time, s.uptime,
                           PRSSM_STATE(s.PRSSM_state));
                    printf("  if_index: %d, name: %s, up: %s, vlan_filter: %s\n",
                           s.if_index, s.name, BOOL_STR(s.up),
                           BOOL_STR(s.vlan_filter));
                  }
              }
            break;
        case PARAM_ENABLED:
            if (use_json && s.enabled)
              {
                json_object_boolean_true_add(json_bridge, "paramEnabled");
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%s\n", BOOL_STR(s.enabled));
            break;
        case PARAM_BRID:
            if (use_json)
              {
                char tempA[BUFSIZ];
                sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.bridge_id));
                json_object_string_add(json_bridge, "paramBridgeId", tempA);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf(BR_ID_FMT"\n", BR_ID_ARGS(s.bridge_id));
            break;
        case PARAM_DSGNROOT:
            if (use_json)
              {
                char tempA[BUFSIZ];
                sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.designated_root));
                json_object_string_add(json_bridge, "paramDsgnRoot", tempA);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf(BR_ID_FMT"\n", BR_ID_ARGS(s.designated_root));
            break;
        case PARAM_REGNROOT:
            if (use_json)
              {
                char tempA[BUFSIZ];
                sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.regional_root));
                json_object_string_add(json_bridge, "paramRegnRoot", tempA);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf(BR_ID_FMT"\n", BR_ID_ARGS(s.regional_root));
            break;
        case PARAM_ROOTPORT:
            if(0 != (root_portno = GET_NUM_FROM_PRIO(s.root_port_id)))
              {
                if (use_json)
                  {
                    json_object_string_add(json_bridge, "paramRootPort", root_port_name);
                    json_object_object_add(json, br_name, json_bridge);
                  }
                else
                  printf("%s\n", root_port_name);
              }
            else
              {
                if (!use_json)
                  printf("\n");
              }
            break;
        case PARAM_PATHCOST:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramPathCost", s.root_path_cost);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%u\n", s.root_path_cost);
            break;
        case PARAM_INTPATHCOST:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramIntPathCost", s.internal_path_cost);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%u\n", s.internal_path_cost);
            break;
        case PARAM_MAXAGE:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramRootMaxAge", s.root_max_age);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%hhu\n", s.root_max_age);
            break;
        case PARAM_BRMAXAGE:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramBridgeMaxAge", s.bridge_max_age);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%hhu\n", s.bridge_max_age);
            break;
        case PARAM_FWDDELAY:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramRootFwdDelay", s.root_forward_delay);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%hhu\n", s.root_forward_delay);
            break;
        case PARAM_BRFWDDELAY:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramBridgeFwdDelay", s.bridge_forward_delay);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%hhu\n", s.bridge_forward_delay);
            break;
        case PARAM_TXHOLDCNT:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramTxHoldCounter", s.tx_hold_count);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%u\n", s.tx_hold_count);
            break;
        case PARAM_MAXHOPS:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramMaxHops", s.max_hops);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%hhu\n", s.max_hops);
            break;
        case PARAM_FORCEPROTVERS:
            if (use_json)
              {
                json_object_string_add(json_bridge, "paramProtocolVersion", PROTO_VERS_STR(s.protocol_version));
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%s\n", PROTO_VERS_STR(s.protocol_version));
            break;
        case PARAM_TOPCHNGTIME:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramTimeSinceTopoChange", s.time_since_topology_change);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%u\n", s.time_since_topology_change);
            break;
        case PARAM_TOPCHNGCNT:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramTopoChngCounter", s.topology_change_count);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%u\n", s.topology_change_count);
            break;
        case PARAM_TOPCHNGSTATE:
            if (use_json && s.topology_change)
              {
                json_object_boolean_true_add(json_bridge, "paramTopoChng");
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%s\n", BOOL_STR(s.topology_change));
            break;
        case PARAM_BRHELLO:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramBridgeHelloTime", s.bridge_hello_time);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%u\n", s.bridge_hello_time);
            break;
        case PARAM_BRAGEING:
            if (use_json)
              {
                json_object_int_add(json_bridge, "paramAgeingTime", s.Ageing_Time);
                json_object_object_add(json, br_name, json_bridge);
              }
            else
              printf("%u\n", s.Ageing_Time);
            break;
        default:
            return -2; /* -2 = unknown param */
    }
    return 0;
}

static int showglobal(u_char use_json, json_object *json)
{
    GlobalStatus s;

    if(CTL_get_global_status(&s))
        return -1;
    if (!use_json) {
        printf("\nGlobal info \n");
        printf("  debug level       %d\n", s.debug_level);
    }

    return 0;
}

static int isbridge(const struct dirent *entry)
{
    char path[SYSFS_PATH_MAX];
    int save_errno;
    bool result;
    struct stat st;
    FILE *fd;
    unsigned char stp_state;

    snprintf(path, SYSFS_PATH_MAX, SYSFS_CLASS_NET "/%s/bridge",
             entry->d_name);
    save_errno = errno;
    result = (0 == stat(path, &st)) && S_ISDIR(st.st_mode);
    errno = save_errno;
    if (result) {
        strcat(path, "/stp_state");
        if ((0 == stat(path, &st)) && (fd = fopen(path, "r"))) {
            fread(&stp_state, 1, 1, fd);
            fclose(fd);
            if (((unsigned int)stp_state - '0') != 2) {
                return 0;
            }
        } else
            result = 0;
    }

    return result;
}

static int cmd_showbridge(int argc, char *const *argv)
{
    int i, count = 0;
    int r = 0;
    struct dirent **namelist;
    param_id_t param_id = PARAM_NULL;
    u_char use_json = 0;
    json_object *json = NULL;

    if (argc > 2)
      {
        if (strcmp(argv[1], "json") == 0)
          {
            use_json = 1;
            json = json_object_new_object();
            count = argc - 2;
          }
        else
          {
            use_json = 0;
            count = argc - 1;
          }

        if (count > 1)
          { /* check if last argument is known parameter name */
            for (i = 0; i < COUNT_OF(cist_bridge_params); ++i)
              {
                if (0 == strcmp(argv[count], cist_bridge_params[i].str))
                  {
                    param_id = cist_bridge_params[i].id;
                    --count;
                    break;
                  }
              }
          }
      }
    else if (1 < argc)
      {
        if (strcmp(argv[1], "json") == 0)
          {
            use_json = 1;
            json = json_object_new_object();
            count = scandir(SYSFS_CLASS_NET, &namelist, isbridge, sorting_func);
            if (0 > count)
              {
                json_object_string_add(json, "error", "Getting list of all bridges");
                printf("\n%s\n", json_object_to_json_string(json));
                json_object_free(json);
                return -1;
              }
          }
        else
          {
            count = argc - 1;
            if (1 < count)
              { /* check if last argument is known parameter name */
                for (i = 0; i < COUNT_OF(cist_bridge_params); ++i)
                  {
                    if (0 == strcmp(argv[count], cist_bridge_params[i].str))
                      {
                        param_id = cist_bridge_params[i].id;
                        --count;
                        break;
                      }
                  }
              }
          }
      }
    else
      {
        count = scandir(SYSFS_CLASS_NET, &namelist, isbridge, sorting_func);
        if (0 > count)
          {
            fprintf(stderr, "Error getting list of all bridges\n");
            return -1;
          }
      }

    for(i = 0; i < count; ++i)
      {
        const char *name;
        if (2 < argc)
          {
            if (strcmp(argv[1], "json") == 0)
              name = argv[i + 2];
            else
              name = argv[i + 1];
          }
        else if (1 < argc)
          {
            if (strcmp(argv[1], "json") == 0)
              name = namelist[i]->d_name;
            else
              name = argv[i + 1];
          }
        else
          name = namelist[i]->d_name;

        int err = do_showbridge(name, param_id, use_json, json);
        if(err)
            r = err;
      }

    if(1 >= argc)
      {
        for(i = 0; i < count; ++i)
          free(namelist[i]);
        free(namelist);
      }

    if (use_json)
      {
          printf("\n%s\n", json_object_to_json_string(json));
          json_object_free(json);
      }

    return r;
}

static int cmd_showtree(int argc, char *const *argv)
{
    MSTI_BridgeStatus s;
    char root_port_name[IFNAMSIZ];
    unsigned int root_portno;
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    if (is_stp_userspace_bridge(argv[1]) == 0)
        return 0;
    int mstid = get_id(argv[2], "mstid", MAX_MSTID);
    if(0 > mstid)
        return mstid;

    if(CTL_get_msti_bridge_status(br_index, mstid, &s, root_port_name))
        return -1;

    printf("%s MSTI %hu info\n", argv[1], mstid);
    printf("  bridge id          "BR_ID_FMT"\n", BR_ID_ARGS(s.bridge_id));
    printf("  regional root      "BR_ID_FMT"\n", BR_ID_ARGS(s.regional_root));
    printf("  root port          ");
    if(0 != (root_portno = GET_NUM_FROM_PRIO(s.root_port_id)))
        printf("%s (#%u)\n", root_port_name, root_portno);
    else
        printf("none\n");
    printf("  internal path cost %u\n", s.internal_path_cost);
    printf("  time since topology change %u\n", s.time_since_topology_change);
    printf("  topology change count      %u\n", s.topology_change_count);
    printf("  topology change            %s\n", BOOL_STR(s.topology_change));
    printf("  topology change port       %s\n", s.topology_change_port);
    printf("  last topology change port  %s\n", s.last_topology_change_port);

    return 0;
}

#define STATE_STR(_state)                                        \
    ({                                                           \
        int _s = _state;                                         \
        char *_str = "unknown";                                  \
        switch(_s)                                               \
        {                                                        \
            case BR_STATE_DISABLED:                              \
            case BR_STATE_BLOCKING:                              \
            case BR_STATE_LISTENING: _str = "discarding"; break; \
            case BR_STATE_LEARNING:  _str = "learning"; break;   \
            case BR_STATE_FORWARDING:_str = "forwarding"; break; \
        }                                                        \
        _str;                                                    \
    })

#define SHORT_STATE_STR(_state)                            \
    ({                                                     \
        int _s = _state;                                   \
        char *_str = "unkn";                               \
        switch(_s)                                         \
        {                                                  \
            case BR_STATE_DISABLED:                        \
            case BR_STATE_BLOCKING:                        \
            case BR_STATE_LISTENING: _str = "disc"; break; \
            case BR_STATE_LEARNING:  _str = "lear"; break; \
            case BR_STATE_FORWARDING:_str = "forw"; break; \
        }                                                  \
        _str;                                              \
    })

#define ADMIN_P2P_STR(_state)                        \
    ({                                               \
        admin_p2p_t _s = _state;                     \
        char *_str = "unkn";                         \
        switch(_s)                                   \
        {                                            \
            case p2pForceFalse:_str = "no"; break;   \
            case p2pForceTrue: _str = "yes"; break;  \
            case p2pAuto:      _str = "auto"; break; \
        }                                            \
        _str;                                        \
    })

#define PORT_INFO(_port_info)                                      \
    ({                                                          \
        port_info_t _p = _port_info;                              \
        char *_str = "unknown";                                 \
        switch(_p)                                              \
        {                                                       \
            case OtherInfo:         _str = "OtherInfo"; break;       \
            case SuperiorDesignatedInfo: _str = "SuperiorDesignatedInfo";break;\
            case RepeatedDesignatedInfo: _str = "RepeatedDesignatedInfo";break;\
            case InferiorDesignatedInfo: _str = "InferiorDesignatedInfo";break;\
            case InferiorRootAlternateInfo:_str = "InferiorRootAlternateInfo";\
                                                                    break;\
        }                                                       \
        _str;                                                   \
    })

#define PORT_INFO_ORIGIN(_port_info_origin)                         \
    ({                                                              \
        port_info_origin_t _p = _port_info_origin;                   \
        char *_str = "unknown";                                     \
        switch(_p)                                                  \
        {                                                           \
            case ioDisabled:         _str = "ioDisabled"; break;    \
            case ioMine:            _str = "ioMine";break;          \
            case ioAged:            _str = "ioAged";break;          \
            case ioReceived:        _str = "ioReceived";break;      \
        }                                                           \
        _str;                                                       \
    })



static const cmd_param_t cist_port_params[] = {
    { PARAM_ENABLED,      "enabled" },
    { PARAM_ROLE,         "role" },
    { PARAM_STATE,        "state" },
    { PARAM_PORTID,       "port-id" },
    { PARAM_EXTPORTCOST,  "external-port-cost" },
    { PARAM_ADMINEXTCOST, "admin-external-cost" },
    { PARAM_INTPORTCOST,  "internal-port-cost" },
    { PARAM_ADMININTCOST, "admin-internal-cost" },
    { PARAM_DSGNROOT,     "designated-root" },
    { PARAM_DSGNEXTCOST,  "dsgn-external-cost" },
    { PARAM_DSGNRROOT,    "dsgn-regional-root" },
    { PARAM_DSGNINTCOST,  "dsgn-internal-cost" },
    { PARAM_DSGNBR,       "designated-bridge" },
    { PARAM_DSGNPORT,     "designated-port" },
    { PARAM_ADMINEDGEPORT,"admin-edge-port" },
    { PARAM_AUTOEDGEPORT, "auto-edge-port" },
    { PARAM_OPEREDGEPORT, "oper-edge-port" },
    { PARAM_TOPCHNGACK,   "topology-change-ack" },
    { PARAM_P2P,          "point-to-point" },
    { PARAM_ADMINP2P,     "admin-point-to-point" },
    { PARAM_RESTRROLE,    "restricted-role" },
    { PARAM_RESTRTCN,     "restricted-TCN" },
    { PARAM_PORTHELLOTIME,"port-hello-time" },
    { PARAM_DISPUTED,     "disputed" },
    { PARAM_BPDUGUARDPORT, "bpdu-guard-port" },
    { PARAM_BPDUGUARDERROR, "bpdu-guard-error" },
    { PARAM_NETWORKPORT, "network-port" },
    { PARAM_BA_INCONSISTENT, "ba-inconsistent" },
    { PARAM_BPDUFILTERPORT, "bpdufilter-port" },
};

static int detail = 0;

static int do_showport(int br_index, const char *bridge_name, const char *port_name,
                       param_id_t param_id, u_char use_json, json_object *json_port)
{
    CIST_PortStatus s;
    json_object *json_name = NULL;
    json_object *json_fail = NULL;
    json_object *json_all = NULL;
    int r = 0;
    int port_index = get_index_die(port_name, "port", false, use_json);
    unsigned int vid;
    bool vlan_set = false;
    unsigned int prev_vid = 0;
    bool range_set = false;

    if(0 > port_index)
      return port_index;

    if (use_json)
      json_name = json_object_new_object();

    if((r = CTL_get_cist_port_status(br_index, port_index, &s)))
      {
        if (use_json)
          {
            json_fail = json_object_new_object();
            json_object_string_add(json_fail, bridge_name, port_name);
            json_object_object_add(json_port, "portStateRetrievalFailure", json_fail);
          }
        else
          fprintf(stderr, "%s:%s Failed to get port state\n", bridge_name, port_name);
        return -1;
      }

    switch(param_id)
      {
        case PARAM_NULL:
            if(detail)
              {
                char tempA[BUFSIZ];
                if (use_json)
                  {
                    json_all = json_object_new_object();
                    json_object_string_add(json_name, "bridgeName", bridge_name);
                    json_object_string_add(json_name, "portName", port_name);
                    if (s.enabled)
                      json_object_boolean_true_add(json_name, "enabled");
                    json_object_string_add(json_name, "role", ROLE_STR(s.role));
                    json_object_string_add(json_name, "state", STATE_STR(s.state));
                    json_object_int_add(json_name, "extPortCost", s.external_port_path_cost);
                    json_object_int_add(json_name, "adminExtPortCost", s.admin_external_port_path_cost);
                    json_object_int_add(json_name, "intPortCost", s.internal_port_path_cost);
                    json_object_int_add(json_name, "adminIntPortCost", s.admin_internal_port_path_cost);

                    sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.designated_root));
                    json_object_string_add(json_name, "dsgnRoot", tempA);

                    json_object_int_add(json_name, "dsgnExtCost", s.designated_external_cost);

                    sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.designated_regional_root));
                    json_object_string_add(json_name, "dsgnRegRoot", tempA);

                    json_object_int_add(json_name, "dsgnIntCost", s.designated_internal_cost);

                    sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.designated_bridge));
                    json_object_string_add(json_name, "dsgnBr", tempA);

                    sprintf(tempA, "%01hhX.%03hX", PRT_ID_ARGS(s.designated_port));
                    json_object_string_add(json_name, "dsgnPort", tempA);

                    if (s.admin_edge_port)
                      json_object_boolean_true_add(json_name, "adminEdgePort");

                    if (s.auto_edge_port)
                      json_object_boolean_true_add(json_name, "autoEdgePort");

                    if (s.oper_edge_port)
                      json_object_boolean_true_add(json_name, "operEdgePort");

                    if (s.tc_ack)
                      json_object_boolean_true_add(json_name, "topoChngAck");

                    if (s.oper_p2p)
                      json_object_boolean_true_add(json_name, "pointToPoint");

                    json_object_string_add(json_name, "adminPointToPoint", ADMIN_P2P_STR(s.admin_p2p));
                    if (s.restricted_role)
                      json_object_boolean_true_add(json_name, "restrictedRole");

                    if (s.restricted_tcn)
                      json_object_boolean_true_add(json_name, "restrictedTCN");

                    json_object_int_add(json_name, "portHelloTime", s.port_hello_time);
                    if (s.disputed)
                      json_object_boolean_true_add(json_name, "disputed");

                    if (s.bpdu_guard_port)
                      json_object_boolean_true_add(json_name, "bpduGuardPort");

                    if (s.bpdu_guard_error)
                      json_object_boolean_true_add(json_name, "bpduGuardError");

                    if (s.network_port)
                      json_object_boolean_true_add(json_name, "networkPort");

                    if (s.ba_inconsistent)
                      json_object_boolean_true_add(json_name, "baInconsistent");

                    json_object_int_add(json_name, "numTxBpdu", s.num_tx_bpdu);
                    json_object_int_add(json_name, "numTxTcn", s.num_tx_tcn);
                    json_object_int_add(json_name, "numRxBpdu", s.num_rx_bpdu);
                    json_object_int_add(json_name, "numRxTcn", s.num_rx_tcn);
                    json_object_int_add(json_name, "numTransFwd", s.num_trans_fwd);
                    json_object_int_add(json_name, "numTransBlk", s.num_trans_blk);
                    if (s.bpdu_filter_port)
                      json_object_boolean_true_add(json_name, "bpduFilterPort");

                    if (s.clagIsl)
                      json_object_boolean_true_add(json_name, "clagIsl");

                    if (s.clagIslOperUp)
                      json_object_boolean_true_add(json_name, "clagIslOperUp");

                    json_object_string_add(json_name, "clagRole", CLAG_ROLE_STR(s.clagRole));

                    char tempB[BUFSIZ];
                    sprintf(tempB, "%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhx",
                            s.clagDualSrcAddr[0], s.clagDualSrcAddr[1],
                            s.clagDualSrcAddr[2], s.clagDualSrcAddr[3],
                            s.clagDualSrcAddr[4], s.clagDualSrcAddr[5]);
                    json_object_string_add(json_name, "clagDualConnMac", tempB);

                    sprintf(tempA, "%01hhX.%03hX", PRT_ID_ARGS(s.clagRemotePortId));
                    json_object_string_add(json_name, "clagRemotePortId", tempA);

                    sprintf(tempB, "%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX",
                            s.clagSysMac[0], s.clagSysMac[1], s.clagSysMac[2],
                            s.clagSysMac[3], s.clagSysMac[4], s.clagSysMac[5]);
                    json_object_string_add(json_name, "clagSystemMac", tempB);
                  }
                else
                  {
                    printf("%s:%s CIST info\n", bridge_name, port_name);
                    printf("  enabled            %-23s ", BOOL_STR(s.enabled));
                    printf("role                 %s\n", ROLE_STR(s.role));
                    printf("  port id            "PRT_ID_FMT"                   ",
                           PRT_ID_ARGS(s.port_id));
                    printf("state                %s\n", STATE_STR(s.state));
                    printf("  external port cost %-23u ",
                           s.external_port_path_cost);
                    printf("admin external cost  %u\n",
                           s.admin_external_port_path_cost);
                    printf("  internal port cost %-23u ",
                           s.internal_port_path_cost);
                    printf("admin internal cost  %u\n",
                           s.admin_internal_port_path_cost);
                    printf("  designated root    "BR_ID_FMT" ",
                           BR_ID_ARGS(s.designated_root));
                    printf("dsgn external cost   %u\n",
                           s.designated_external_cost);
                    printf("  dsgn regional root "BR_ID_FMT" ",
                           BR_ID_ARGS(s.designated_regional_root));
                    printf("dsgn internal cost   %u\n",
                           s.designated_internal_cost);
                    printf("  designated bridge  "BR_ID_FMT" ",
                           BR_ID_ARGS(s.designated_bridge));
                    printf("designated port      "PRT_ID_FMT"\n",
                           PRT_ID_ARGS(s.designated_port));
                    printf("  admin edge port    %-23s ",
                           BOOL_STR(s.admin_edge_port));
                    printf("auto edge port       %s\n",
                           BOOL_STR(s.auto_edge_port));
                    printf("  oper edge port     %-23s ",
                           BOOL_STR(s.oper_edge_port));
                    printf("topology change ack  %s\n", BOOL_STR(s.tc_ack));
                    printf("  point-to-point     %-23s ", BOOL_STR(s.oper_p2p));
                    printf("admin point-to-point %s\n",
                           ADMIN_P2P_STR(s.admin_p2p));
                    printf("  restricted role    %-23s ",
                           BOOL_STR(s.restricted_role));
                    printf("restricted TCN       %s\n",
                           BOOL_STR(s.restricted_tcn));
                    printf("  port hello time    %-23hhu ", s.port_hello_time);
                    printf("disputed             %s\n", BOOL_STR(s.disputed));
                    printf("  bpdu guard port    %-23s ",
                           BOOL_STR(s.bpdu_guard_port));
                    printf("bpdu guard error     %s\n",
                           BOOL_STR(s.bpdu_guard_error));
                    printf("  network port       %-23s ",
                           BOOL_STR(s.network_port));
                    printf("BA inconsistent      %s\n",
                           BOOL_STR(s.ba_inconsistent));
                    printf("  Num TX BPDU        %-23u ",s.num_tx_bpdu);
                    printf("Num TX TCN           %u\n", s.num_tx_tcn);
                    printf("  Num RX BPDU        %-23u ", s.num_rx_bpdu);
                    printf("Num RX TCN           %u\n", s.num_rx_tcn);
                    printf("  Num Transition FWD %-23u ", s.num_trans_fwd);
                    printf("Num Transition BLK   %u\n", s.num_trans_blk);
                    printf("  bpdufilter port    %-23s\n",
                           BOOL_STR(s.bpdu_filter_port));
                    printf("  clag ISL           %-23s ", BOOL_STR(s.clagIsl));
                    printf("clag ISL Oper UP     %s\n", BOOL_STR(s.clagIslOperUp));
                    printf("  clag role          %-23s ",CLAG_ROLE_STR(s.clagRole));
                    printf("clag dual conn mac   %02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX\n",
                           s.clagDualSrcAddr[0], s.clagDualSrcAddr[1],
                           s.clagDualSrcAddr[2], s.clagDualSrcAddr[3],
                           s.clagDualSrcAddr[4], s.clagDualSrcAddr[5]);
                    printf("  clag remote portID "PRT_ID_FMT"                   ",
                           PRT_ID_ARGS(s.clagRemotePortId));
                    printf("clag system mac      %02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX\n",
                           s.clagSysMac[0], s.clagSysMac[1], s.clagSysMac[2],
                           s.clagSysMac[3], s.clagSysMac[4], s.clagSysMac[5]);
                  }

                if (all)
                  {
                    if (use_json)
                      {
                        if (s.agree)
                          json_object_boolean_true_add(json_all, "agree");
                        if (s.agreed)
                          json_object_boolean_true_add(json_all, "agreed");
                        if (s.forward)
                          json_object_boolean_true_add(json_all, "forward");
                        if (s.forwarding)
                          json_object_boolean_true_add(json_all, "forwarding");
                        if (s.learn)
                          json_object_boolean_true_add(json_all, "learn");
                        if (s.learning)
                          json_object_boolean_true_add(json_all, "learning");
                        if (s.proposed)
                          json_object_boolean_true_add(json_all, "proposed");
                        if (s.proposing)
                          json_object_boolean_true_add(json_all, "proposing");
                        if (s.rcvdMsg)
                          json_object_boolean_true_add(json_all, "rcvdMsg");
                        if (s.rcvdTc)
                          json_object_boolean_true_add(json_all, "rcvdTc");
                        if (s.reRoot)
                          json_object_boolean_true_add(json_all, "reboot");
                        if (s.reselect)
                          json_object_boolean_true_add(json_all, "reselect");
                        if (s.selected)
                          json_object_boolean_true_add(json_all, "selected");
                        if (s.fdbFlush)
                          json_object_boolean_true_add(json_all, "fdbFlush");
                        if (s.tcProp)
                          json_object_boolean_true_add(json_all, "tcProp");
                        if (s.updtInfo)
                          json_object_boolean_true_add(json_all, "updtInfo");
                        if (s.sync)
                          json_object_boolean_true_add(json_all, "sync");
                        if (s.synced)
                          json_object_boolean_true_add(json_all, "synced");
                        if (s.master)
                          json_object_boolean_true_add(json_all, "master");
                        if (s.mastered)
                          json_object_boolean_true_add(json_all, "mastered");
                        json_object_int_add(json_all, "brAssuRcvdInfoWhile", s.brAssuRcvdInfoWhile);
                        if (s.calledFromFlushRoutine)
                          json_object_boolean_true_add(json_all, "calledFromFlushRoutine");
                        if (s.infoInternal)
                          json_object_boolean_true_add(json_all, "infoInternal");
                        if (s.rcvdTcAck)
                          json_object_boolean_true_add(json_all, "rcvdTcAck");
                        if (s.rcvdInternal)
                          json_object_boolean_true_add(json_all, "rcvdInterval");
                        if (s.mcheck)
                          json_object_boolean_true_add(json_all, "mcheck");
                        if (s.rcvdBpdu)
                          json_object_boolean_true_add(json_all, "rcvdBpdu");
                        if (s.rcvdSTP)
                          json_object_boolean_true_add(json_all, "rcvdStp");
                        if (s.rcvdRSTP)
                          json_object_boolean_true_add(json_all, "rcvdRstp");
                        if (s.rcvdTcn)
                          json_object_boolean_true_add(json_all, "rcvdTcn");
                        if (s.sendRSTP)
                          json_object_boolean_true_add(json_all, "sendRstp");
                        if (s.newInfo)
                          json_object_boolean_true_add(json_all, "newInfo");
                        if (s.newInfoMsti)
                          json_object_boolean_true_add(json_all, "newInfoMsti");
                        if (s.up)
                          json_object_boolean_true_add(json_all, "up");
                        json_object_int_add(json_all, "fdWhile", s.fdWhile);
                        json_object_int_add(json_all, "rrWhile", s.rrWhile);
                        json_object_int_add(json_all, "rbWhile", s.rbWhile);
                        json_object_int_add(json_all, "tcWhile", s.tcWhile);
                        json_object_int_add(json_all, "rcvdInfoWhile", s.rcvdInfoWhile);
                        json_object_string_add(json_all, "rcvdInfo", PORT_INFO(s.rcvdInfo));
                        json_object_string_add(json_all, "infoIs", PORT_INFO_ORIGIN(s.infoIs));
                        json_object_string_add(json_all, "selectedRole", ROLE_STR(s.selectedRole));
                        json_object_int_add(json_all, "edgeDelayWhile", s.edgeDelayWhile);
                        json_object_int_add(json_all, "txCount", s.txCount);
                        json_object_int_add(json_all, "rapidAgeingWhile", s.rapidAgeingWhile);
                        json_object_int_add(json_all, "speed", s.speed);
                        json_object_int_add(json_all, "duplex", s.duplex);
                        json_object_string_add(json_all, "portInfoSM", PISM_STATE(s.PISM_state));
                        json_object_string_add(json_all, "portRoleTransititionSM", PRTSM_STATE(s.PRTSM_state));
                        json_object_string_add(json_all, "portStateTransitionSM", PSTSM_STATE(s.PSTSM_state));
                        json_object_string_add(json_all, "topoChangeSM", TCSM_STATE(s.TCSM_state));
                        json_object_string_add(json_all, "portReceiveSM", PRSM_STATE(s.PRSM_state));
                        json_object_string_add(json_all, "portTransmitSM", PTSM_STATE(s.PTSM_state));
                        json_object_string_add(json_all, "portProtocolMigrationSM", PPMSM_STATE(s.PPMSM_state));
                        json_object_string_add(json_all, "bridgeDetectionSM", BDSM_STATE(s.BDSM_state));
                        json_object_int_add(json_all, "ifIndex", s.if_index);
                        json_object_string_add(json_all, "name", s.name);
                        json_object_int_add(json_all, "mdelayWhile", s.mdelayWhile);
                        json_object_int_add(json_all, "helloWhen", s.helloWhen);
                        json_object_string_add(json_all, "vlanBond", s.vlan_bond_dev);
                        json_object_int_add(json_name, "numTxPvstTnlBpdu", s.num_tx_pvst_tnl_bpdu);
                        json_object_int_add(json_name, "numRxPvstTnlBpdu", s.num_rx_pvst_tnl_bpdu);
                        json_object_object_add(json_name, "allInformation", json_all);
                      }
                    else
                      {
                        printf("  agree: %s, agreed: %s, forward: %s, forwarding: %s\n",
                               BOOL_STR(s.agree), BOOL_STR(s.agreed),
                               BOOL_STR(s.forward), BOOL_STR(s.forwarding));
                        printf("  learn: %s, learning: %s, proposed: %s, proposing: %s\n",
                               BOOL_STR(s.learn), BOOL_STR(s.learning),
                               BOOL_STR(s.proposed), BOOL_STR(s.proposing));
                        printf("  rcvdMsg: %s, rcvdTc: %s, reRoot: %s, reselect: %s\n",
                               BOOL_STR(s.rcvdMsg), BOOL_STR(s.rcvdTc),
                               BOOL_STR(s.reRoot), BOOL_STR(s.reselect));
                        printf("  selected: %s, fdbFlush: %s, tcProp: %s, updtInfo: %s\n",
                               BOOL_STR(s.selected), BOOL_STR(s.fdbFlush),
                               BOOL_STR(s.tcProp), BOOL_STR(s.updtInfo));
                        printf("  sync: %s, synced: %s, master: %s, mastered: %s"\
                               " brAssuRcvdInfoWhile: %us\n",
                               BOOL_STR(s.sync), BOOL_STR(s.synced),
                               BOOL_STR(s.master), BOOL_STR(s.mastered),
                               s.brAssuRcvdInfoWhile);
                        printf("  calledFromFlushRoutine: %s, infoInternal: %s," \
                               " rcvdTcAck: %s\n",
                               BOOL_STR(s.calledFromFlushRoutine),
                               BOOL_STR(s.infoInternal),
                               BOOL_STR(s.rcvdTcAck));
                        printf("  rcvdInternal: %s, mcheck: %s, rcvdBpdu: %s," \
                               " rcvdSTP: %s, rcvdRSTP: %s\n",
                               BOOL_STR(s.rcvdInternal), BOOL_STR(s.mcheck),
                               BOOL_STR(s.rcvdBpdu), BOOL_STR(s.rcvdSTP),
                               BOOL_STR(s.rcvdRSTP));
                        printf("  rcvdTcn: %s, sendRSTP: %s, newInfo: %s, " \
                               "newInfoMsti: %s, up: %s\n",
                               BOOL_STR(s.rcvdTcn), BOOL_STR(s.sendRSTP),
                               BOOL_STR(s.newInfo), BOOL_STR(s.newInfoMsti),
                               BOOL_STR(s.up));
                        printf("  fdWhile: %us, rrWhile: %us, rbWhile: %u, " \
                               "tcwhile: %us, rcvdInfoWhile: %us\n",
                               s.fdWhile, s.rrWhile, s.rbWhile, s.tcWhile,
                               s.rcvdInfoWhile);
                        printf("  rcvdInfo: %s, infoIs: %s, selectedRole: %s\n",
                               PORT_INFO(s.rcvdInfo), PORT_INFO_ORIGIN(s.infoIs),
                               ROLE_STR(s.selectedRole));
                        printf("  edgeDelayWhile: %us, txCount: %u," \
                               " rapidAgeingWhile: %us," \
                               " speed: %u, duplex: %u\n",
                               s.edgeDelayWhile, s.txCount, s.rapidAgeingWhile, s.speed,
                               s.duplex);
                        printf("  Port Info SM: %s, Port Role Transitition SM: %s\n",
                               PISM_STATE(s.PISM_state), PRTSM_STATE(s.PRTSM_state));
                        printf("  Port State Transition SM: %s, Topo Change SM: %s\n",
                               PSTSM_STATE(s.PSTSM_state), TCSM_STATE(s.TCSM_state));
                        printf("  Port Receive SM: %s, Port Transmit SM: %s\n",
                               PRSM_STATE(s.PRSM_state), PTSM_STATE(s.PTSM_state));
                        printf("  Port Protocol Migration SM: %s, "\
                               "Bridge Detection SM: %s\n",
                               PPMSM_STATE(s.PPMSM_state), BDSM_STATE(s.BDSM_state));
                        printf("  if_index: %d, name: %s, mdelayWhile: %us, " \
                               "helloWhen: %us, vlan_bond: %s\n",
                               s.if_index, s.name, s.mdelayWhile, s.helloWhen,
                               s.vlan_bond_dev);
                        printf("  numTxPvstTnlBpdu: %u, numRxPvstTnlBpdu: %u\n",
                                s.num_tx_pvst_tnl_bpdu, s.num_rx_pvst_tnl_bpdu);

                        for_each_set_bit(vid, s.vlan_bitmap, VLAN_N_VID) {
                            if (!vlan_set) {
                                printf("  vlan: %d", vid);
                                vlan_set = true;
                                prev_vid = vid;
                                continue;
                            }
                            if (prev_vid) {
                                if ((prev_vid+1) != vid) {
                                    if (range_set) {
                                        printf("%d, %d", prev_vid, vid);
                                        range_set = false;
                                    } else
                                        printf(", %d", vid);
                                } else {
                                    if (!range_set) {
                                        range_set = true;
                                        printf("-");
                                    }
                                }
                                prev_vid = vid;
                            }
                        }
                        if (vlan_set) {
                            if (range_set)
                                printf("%d", prev_vid);
                            printf("\n");
                        }
                        printf("\n");
                      }
                  }
                if (use_json)
                  {
                    sprintf(tempA, "%01hhX.%03hX", PRT_ID_ARGS(s.port_id));
                    json_object_object_add(json_port, tempA, json_name);
                  }
              }
            else
              {
                if (use_json)
                  {
                    if (s.oper_p2p)
                      json_object_boolean_true_add(json_name, "operPeerToPeer");
                    if (s.oper_edge_port)
                      json_object_boolean_true_add(json_name, "operEdgePort");
                    json_object_string_add(json_name, "portName", port_name);
                    if (s.enabled)
                      json_object_string_add(json_name, "state", SHORT_STATE_STR(s.state));
                    else
                      json_object_string_add(json_name, "state", "down");

                    char tempA[BUFSIZ];
                    sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.designated_root));
                    json_object_string_add(json_name, "dsgnRoot", tempA);

                    sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.designated_bridge));
                    json_object_string_add(json_name, "dsgnBr", tempA);

                    sprintf(tempA, "%01hhX.%03hX", PRT_ID_ARGS(s.designated_port));
                    json_object_string_add(json_name, "dsgnPort", tempA);

                    json_object_string_add(json_name, "role", SHORT_ROLE_STR(s.role));

                    sprintf(tempA, "%01hhX.%03hX", PRT_ID_ARGS(s.port_id));
                    json_object_object_add(json_port, tempA, json_name);
                  }
                else
                  printf("%c%c %-5s "PRT_ID_FMT" %4s "BR_ID_FMT" "BR_ID_FMT" "PRT_ID_FMT" %s\n",
                         (s.oper_p2p) ? ' ' : '*',
                         (s.oper_edge_port) ? 'E' : ' ',
                         port_name,
                         PRT_ID_ARGS(s.port_id),
                         s.enabled ? SHORT_STATE_STR(s.state) : "down",
                         BR_ID_ARGS(s.designated_root),
                         BR_ID_ARGS(s.designated_bridge),
                         PRT_ID_ARGS(s.designated_port),
                         SHORT_ROLE_STR(s.role));
              }
            break;
        case PARAM_ENABLED:
            if (use_json && s.enabled)
              json_object_boolean_true_add(json_port, "paramEnabled");
            else
              printf("%s\n", BOOL_STR(s.enabled));
            break;
        case PARAM_ROLE:
            if (use_json)
              json_object_string_add(json_name, "paramRole", ROLE_STR(s.role));
            else
              printf("%s\n", ROLE_STR(s.role));
            break;
        case PARAM_STATE:
            if (use_json)
              json_object_string_add(json_port, "paramState", STATE_STR(s.state));
            else
              printf("%s\n", STATE_STR(s.state));
            break;
        case PARAM_PORTID:
            if (use_json)
              {
                char tempA[BUFSIZ];
                sprintf(tempA, "%01hhX.%03hX", PRT_ID_ARGS(s.port_id));
                json_object_string_add(json_port, "paramPortId", tempA);
              }
            else
              printf(PRT_ID_FMT"\n", PRT_ID_ARGS(s.port_id));
            break;
        case PARAM_EXTPORTCOST:
            if (use_json)
              json_object_int_add(json_port, "paramExtPortCost", s.external_port_path_cost);
            else
              printf("%u\n", s.external_port_path_cost);
            break;
        case PARAM_ADMINEXTCOST:
            if (use_json)
              json_object_int_add(json_port, "paramAdminExtPortCost", s.admin_external_port_path_cost);
            else
              printf("%u\n", s.admin_external_port_path_cost);
            break;
        case PARAM_INTPORTCOST:
            if (use_json)
              json_object_int_add(json_port, "paramIntPortCost", s.internal_port_path_cost);
            else
              printf("%u\n", s.internal_port_path_cost);
            break;
        case PARAM_ADMININTCOST:
            if (use_json)
              json_object_int_add(json_port, "paramAdminIntPortCost", s.admin_internal_port_path_cost);
            else
              printf("%u\n", s.admin_internal_port_path_cost);
            break;
        case PARAM_DSGNROOT:
            if (use_json)
              {
                char tempA[BUFSIZ];
                sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.designated_root));
                json_object_string_add(json_port, "paramDsgnRoot", tempA);
              }
            else
              printf(BR_ID_FMT"\n", BR_ID_ARGS(s.designated_root));
            break;
        case PARAM_DSGNEXTCOST:
            if (use_json)
              json_object_int_add(json_port, "paramDsgnExtCost", s.designated_external_cost);
            else
              printf("%u\n", s.designated_external_cost);
            break;
        case PARAM_DSGNRROOT:
            if (use_json)
              {
                char tempA[BUFSIZ];
                sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.designated_regional_root));
                json_object_string_add(json_port, "paramDsgnRegRoot", tempA);
              }
            else
              printf(BR_ID_FMT"\n", BR_ID_ARGS(s.designated_regional_root));
            break;
        case PARAM_DSGNINTCOST:
            if (use_json)
              json_object_int_add(json_port, "paramDsgnIntCost", s.designated_internal_cost);
            else
              printf("%u\n", s.designated_internal_cost);
            break;
        case PARAM_DSGNBR:
            if (use_json)
              {
                char tempA[BUFSIZ];
                sprintf(tempA, "%01hhX.%03hX.%02hhX:%02hhX:%02hhX:%02hhX:%02hhX:%02hhX", BR_ID_ARGS(s.designated_bridge));
                json_object_string_add(json_port, "paramDsgnBr", tempA);
              }
            else
              printf(BR_ID_FMT"\n", BR_ID_ARGS(s.designated_bridge));
            break;
        case PARAM_DSGNPORT:
            if (use_json)
              {
                char tempA[BUFSIZ];
                sprintf(tempA, "%01hhX.%03hX", PRT_ID_ARGS(s.designated_port));
                json_object_string_add(json_port, "paramDsgnPort", tempA);
              }
            else
              printf(PRT_ID_FMT"\n", PRT_ID_ARGS(s.designated_port));
            break;
        case PARAM_ADMINEDGEPORT:
            if (use_json && s.admin_edge_port)
              json_object_boolean_true_add(json_port, "paramAdminEdgePort");
            else
              printf("%s\n", BOOL_STR(s.admin_edge_port));
            break;
        case PARAM_AUTOEDGEPORT:
            if (use_json && s.auto_edge_port)
              json_object_boolean_true_add(json_port, "paramAutoEdgePort");
            else
              printf("%s\n", BOOL_STR(s.auto_edge_port));
            break;
        case PARAM_OPEREDGEPORT:
            if (use_json && s.oper_edge_port)
              json_object_boolean_true_add(json_port, "paramOperEdgePort");
            else
              printf("%s\n", BOOL_STR(s.oper_edge_port));
            break;
        case PARAM_TOPCHNGACK:
            if (use_json && s.tc_ack)
              json_object_boolean_true_add(json_port, "paramTopChngAck");
            else
              printf("%s\n", BOOL_STR(s.tc_ack));
            break;
        case PARAM_P2P:
            if (use_json && s.oper_p2p)
              json_object_boolean_true_add(json_port, "paramPeerToPeer");
            else
              printf("%s\n", BOOL_STR(s.oper_p2p));
            break;
        case PARAM_ADMINP2P:
            if (use_json)
              json_object_string_add(json_port, "paramAdminPeerToPeer", ADMIN_P2P_STR(s.admin_p2p));
            else
              printf("%s\n", ADMIN_P2P_STR(s.admin_p2p));
            break;
        case PARAM_RESTRROLE:
            if (use_json && s.restricted_role)
              json_object_boolean_true_add(json_port, "paramRestrictedRole");
            else
              printf("%s\n", BOOL_STR(s.restricted_role));
            break;
        case PARAM_RESTRTCN:
            if (use_json && s.restricted_tcn)
              json_object_boolean_true_add(json_port, "paramRestrictedTCN");
            else
              printf("%s\n", BOOL_STR(s.restricted_tcn));
            break;
        case PARAM_PORTHELLOTIME:
            if (use_json)
              json_object_int_add(json_port, "paramPortHelloTime", s.port_hello_time);
            else
              printf("%hhu\n", s.port_hello_time);
            break;
        case PARAM_DISPUTED:
            if (use_json && s.disputed)
              json_object_boolean_true_add(json_port, "paramDisputed");
            else
              printf("%s\n", BOOL_STR(s.disputed));
            break;
        case PARAM_BPDUGUARDPORT:
            if (use_json && s.bpdu_guard_port)
              json_object_boolean_true_add(json_port, "paramBpduGuardPort");
            else
              printf("%s\n", BOOL_STR(s.bpdu_guard_port));
            break;
        case PARAM_BPDUGUARDERROR:
            if (use_json && s.bpdu_guard_error)
              json_object_boolean_true_add(json_port, "paramBpduGuardError");
            else
              printf("%s\n", BOOL_STR(s.bpdu_guard_error));
            break;
        case PARAM_NETWORKPORT:
            if (use_json && s.network_port)
              json_object_boolean_true_add(json_port, "paramNetworkPort");
            else
              printf("%s\n", BOOL_STR(s.network_port));
            break;
        case PARAM_BA_INCONSISTENT:
            if (use_json && s.ba_inconsistent)
              json_object_boolean_true_add(json_port, "paramBaInconsistent");
            else
              printf("%s\n", BOOL_STR(s.ba_inconsistent));
            break;
        case PARAM_BPDUFILTERPORT:
            if (use_json && s.bpdu_filter_port)
              json_object_boolean_true_add(json_port, "paramBpduFilterPort");
            else
              printf("%s\n", BOOL_STR(s.bpdu_filter_port));
            break;
        default:
            return -2; /* -2 = unknown param */
      }
    return 0;
}

static int not_dot_dotdot(const struct dirent *entry)
{
    const char *n = entry->d_name;

    return !('.' == n[0] && (0 == n[1] || ('.' == n[1] && 0 == n[2])));
}

static int get_port_list(const char *br_ifname, struct dirent ***namelist)
{
    int res;
    char buf[SYSFS_PATH_MAX];

    snprintf(buf, sizeof(buf), SYSFS_CLASS_NET "/%s/brif", br_ifname);
    if(0 > (res = scandir(buf, namelist, not_dot_dotdot, sorting_func)))
        fprintf(stderr, "Error getting list of all ports of bridge %s\n",
                br_ifname);
    return res;
}

static int cmd_showport(int argc, char *const *argv)
{
    int r = 0;
    u_char use_json = 0;
    int br_index = get_index(argv[1], "bridge");
    if (0 > br_index)
      return br_index;
    if (is_stp_userspace_bridge(argv[1]) == 0)
      return 0;

    int i, count = 0;
    struct dirent **namelist;
    param_id_t param_id = PARAM_NULL;
    json_object *json = NULL;

    if (argc > 3)
      {
        if (strcmp(argv[2], "json") == 0)
          {
            use_json = 1;
            json = json_object_new_object();
            count = argc - 3;
          }
        else
          {
            use_json = 0;
            count = argc - 2;
          }
        if (1 < count)
          { /* check if last argument is known parameter name */
            for (i = 0; i < COUNT_OF(cist_port_params); ++i)
              {
                if (0 == strcmp(argv[argc - 1], cist_port_params[i].str))
                  {
                    param_id = cist_port_params[i].id;
                    --count;
                    break;
                  }
              }
          }
      }
    else if (argc > 2)
      {
        if (strcmp(argv[2], "json") == 0)
          {
            use_json = 1;
            json = json_object_new_object();
            if (0 > (count = get_port_list(argv[1], &namelist)))
              return count;
          }
        else
          {
            use_json = 0;
            count = argc - 2;
          }

        if (1 < count)
          { /* check if last argument is known parameter name */
            for (i = 0; i < COUNT_OF(cist_port_params); ++i)
              {
                if (0 == strcmp(argv[argc - 1], cist_port_params[i].str))
                  {
                    param_id = cist_port_params[i].id;
                    --count;
                    break;
                  }
              }
          }
      }
    else
      {
        if (0 > (count = get_port_list(argv[1], &namelist)))
          return count;
      }

    for (i = 0; i < count; ++i)
      {
        json_object *json_port = NULL;

        if (use_json)
          json_port = json_object_new_object();

        const char *name;
        if (argc > 3)
          {
            if (strcmp(argv[2], "json") == 0)
              name = argv[i + 3];
            else
              name = argv[i + 2];
          }
        else if (argc > 2)
          {
            if (strcmp(argv[2], "json") == 0)
              name = namelist[i]->d_name;
            else
              name = argv[i + 2];
          }
        else
          name = namelist[i]->d_name;

        int err = do_showport(br_index, argv[1], name, param_id, use_json, json_port);

        if (use_json)
          json_object_object_add(json, name, json_port);

        if (err)
          r = err;
      }

    if (argc <= 2)
      {
        for (i = 0; i < count; ++i)
          free(namelist[i]);
        free(namelist);
      }

    if (use_json)
      {
        printf("\n%s\n", json_object_to_json_string(json));
        json_object_free(json);
      }

    return r;
}

static int cmd_showportdetail(int argc, char *const *argv)
{
    detail = 1;
    return cmd_showport(argc, argv);
}

static int cmd_showtreeport(int argc, char *const *argv)
{
    MSTI_PortStatus s;
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    if (is_stp_userspace_bridge(argv[1]) == 0)
        return 0;
    int port_index = get_index(argv[2], "port");
    if(0 > port_index)
        return port_index;
    int mstid = get_id(argv[3], "mstid", MAX_MSTID);
    if(0 > mstid)
        return mstid;

    if(CTL_get_msti_port_status(br_index, port_index, mstid, &s))
        return -1;

    printf("%s:%s MSTI %hu info\n", argv[1], argv[2], mstid);
    printf("  role               %-23s ", ROLE_STR(s.role));
    printf("port id              "PRT_ID_FMT"\n", PRT_ID_ARGS(s.port_id));
    printf("  state              %-23s ", STATE_STR(s.state));
    printf("disputed             %s\n", BOOL_STR(s.disputed));
    printf("  internal port cost %-23u ", s.internal_port_path_cost);
    printf("admin internal cost  %u\n", s.admin_internal_port_path_cost);
    printf("  dsgn regional root "BR_ID_FMT" ",
           BR_ID_ARGS(s.designated_regional_root));
    printf("dsgn internal cost   %u\n", s.designated_internal_cost);
    printf("  designated bridge  "BR_ID_FMT" ",
           BR_ID_ARGS(s.designated_bridge));
    printf("designated port      "PRT_ID_FMT"\n",
           PRT_ID_ARGS(s.designated_port));

    return 0;
}

static int cmd_addbridge(int argc, char *const *argv)
{
    int i, j, res, ifcount, brcount = argc - 1;
    int *br_array;
    int* *ifaces_lists;

    if(NULL == (br_array = malloc((brcount + 1) * sizeof(int))))
    {
out_of_memory_exit:
        fprintf(stderr, "out of memory, brcount = %d\n", brcount);
        return -1;
    }
    if(NULL == (ifaces_lists = malloc(brcount * sizeof(int*))))
    {
        free(br_array);
        goto out_of_memory_exit;
    }

    br_array[0] = brcount;
    for(i = 1; i <= brcount; ++i)
    {
        struct dirent **namelist;

        br_array[i] = get_index(argv[i], "bridge");

        if(0 > (ifcount = get_port_list(argv[i], &namelist)))
        {
ifaces_error_exit:
            for(i -= 2; i >= 0; --i)
                free(ifaces_lists[i]);
            free(ifaces_lists);
            free(br_array);
            return ifcount;
        }

        if(NULL == (ifaces_lists[i - 1] = malloc((ifcount + 1) * sizeof(int))))
        {
            fprintf(stderr, "out of memory, bridge %s, ifcount = %d\n",
                    argv[i], ifcount);
            for(j = 0; j < ifcount; ++j)
                free(namelist[j]);
            free(namelist);
            ifcount = -1;
            goto ifaces_error_exit;
        }

        ifaces_lists[i - 1][0] = ifcount;
        for(j = 1; j <= ifcount; ++j)
        {
            ifaces_lists[i - 1][j] = get_index(namelist[j - 1]->d_name, "port");
            free(namelist[j - 1]);
        }
        free(namelist);
    }

    res = CTL_add_bridges(br_array, ifaces_lists);

    for(i = 0; i < brcount; ++i)
        free(ifaces_lists[i]);
    free(ifaces_lists);
    free(br_array);
    return res;
}

static int cmd_delbridge(int argc, char *const *argv)
{
    int i, res, brcount = argc - 1;
    int *br_array;

    if(NULL == (br_array = malloc((brcount + 1) * sizeof(int))))
    {
        fprintf(stderr, "out of memory, brcount = %d\n", brcount);
        return -1;
    }

    br_array[0] = brcount;
    for(i = 1; i <= brcount; ++i)
        br_array[i] = get_index(argv[i], "bridge");

    res = CTL_del_bridges(br_array);

    free(br_array);
    return res;
}

static unsigned int getuint(const char *s)
{
    char *end;
    unsigned long l;
    l = strtoul(s, &end, 0);
    if(0 == *s || 0 != *end || INT_MAX <= l)
    {
        fprintf(stderr, "Invalid unsigned int arg %s\n", s);
        exit(1);
    }
    return l;
}

static int getenum(const char *s, const char *opt[])
{
    int i;
    for(i = 0; opt[i] != NULL; ++i)
        if(0 == strcmp(s, opt[i]))
            return i;

    fprintf(stderr, "Invalid argument %s: expecting one of ", s);
    for(i = 0; opt[i] != NULL; ++i)
        fprintf(stderr, "%s%s", opt[i], (opt[i + 1] ? ", " : "\n"));

    exit(1);
}

static int getyesno(const char *s, const char *yes, const char *no)
{
    /* Reverse yes and no so error message looks more normal */
    const char *opt[] = { yes, no, NULL };
    return 1 - getenum(s, opt);
}

static int cmd_setmstconfid(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    unsigned int revision = getuint(argv[2]);
    if(revision > 0xFFFF)
    {
        fprintf(stderr, "Bad revision %s\n", argv[2]);
        return -1;
    }
    return CTL_set_mstconfid(br_index, revision, argv[3]);
}

#define set_bridge_cfg(field, value)                       \
    ({                                                     \
        CIST_BridgeConfig c;                               \
        memset(&c, 0, sizeof(c));                          \
        c.field = value;                                   \
        c.set_ ## field = true;                            \
        int r = CTL_set_cist_bridge_config(br_index, &c);  \
        if(r)                                              \
            printf("Couldn't change bridge " #field "\n"); \
        r;                                                 \
    })

#define set_port_cfg(field, value)                                  \
    ({                                                              \
        CIST_PortConfig c;                                          \
        memset(&c, 0, sizeof(c));                                   \
        c.field = value;                                            \
        c.set_ ## field = true;                                     \
        int r = CTL_set_cist_port_config(br_index, port_index, &c); \
        if(r)                                                       \
            printf("Couldn't change port " #field "\n");            \
        r;                                                          \
    })

#define set_tree_port_cfg(field, value)                                    \
    ({                                                                     \
        MSTI_PortConfig c;                                                 \
        memset(&c, 0, sizeof(c));                                          \
        c.field = value;                                                   \
        c.set_ ## field = true;                                            \
        int r = CTL_set_msti_port_config(br_index, port_index, mstid, &c); \
        if(r)                                                              \
            printf("Couldn't change per-tree port " #field "\n");          \
        r;                                                                 \
    })

static int cmd_setbridgemaxage(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    unsigned int max_age = getuint(argv[2]);
    if(max_age > 255)
        max_age = 255;
    return set_bridge_cfg(bridge_max_age, max_age);
}

static int cmd_setbridgehello(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    unsigned int hello_time = getuint(argv[2]);
    if(hello_time > 255)
        hello_time = 255;
    return set_bridge_cfg(bridge_hello_time, hello_time);
}

static int cmd_setbridgeageing(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    unsigned int ageing_time = getuint(argv[2]);
    if(ageing_time > 1100000)
        ageing_time = 1100000;
    return set_bridge_cfg(bridge_ageing_time, ageing_time);
}


static int cmd_setbridgefdelay(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    unsigned int forward_delay = getuint(argv[2]);
    if(forward_delay > 255)
        forward_delay = 255;
    return set_bridge_cfg(bridge_forward_delay, forward_delay);
}

static int cmd_setbridgemaxhops(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    unsigned int max_hops = getuint(argv[2]);
    if(max_hops > 255)
        max_hops = 255;
    return set_bridge_cfg(max_hops, max_hops);
}

static int cmd_setbridgeforcevers(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    const char *opts[] = { "stp", "rstp", "mstp", NULL };
    int vals[] = { protoSTP, protoRSTP, protoMSTP };
    return set_bridge_cfg(protocol_version, vals[getenum(argv[2], opts)]);
}

static int cmd_setbridgetxholdcount(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    unsigned int hold_count = getuint(argv[2]);
    if (hold_count > 255)
        hold_count = 255;
    return set_bridge_cfg(tx_hold_count, hold_count);
}

static int cmd_settreeprio(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int mstid = get_id(argv[2], "mstid", MAX_MSTID);
    if(0 > mstid)
        return mstid;
    unsigned int prio = getuint(argv[3]);
    if(prio > 65535)
        prio = 65535;
    return CTL_set_msti_bridge_config(br_index,  mstid, prio);
}

static int cmd_setportpathcost(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if(0 > port_index)
        return port_index;
    unsigned int path_cost = getuint(argv[3]);
    if(path_cost > 210000000)
        path_cost = 210000000;
    return set_port_cfg(admin_external_port_path_cost, getuint(argv[3]));
}

static int cmd_setportadminedge(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if(0 > port_index)
        return port_index;
    return set_port_cfg(admin_edge_port, getyesno(argv[3], "yes", "no"));
}

static int cmd_setportbpduguard(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if (0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if (0 > port_index)
        return port_index;
    return set_port_cfg(bpdu_guard_port, getyesno(argv[3], "yes", "no"));
}

static int cmd_setportnetwork(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if (0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if (0 > port_index)
        return port_index;
    return set_port_cfg(network_port, getyesno(argv[3], "yes", "no"));
}

static int cmd_setportdonttxmt(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if (0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if (0 > port_index)
        return port_index;
    return set_port_cfg(dont_txmt, getyesno(argv[3], "yes", "no"));
}

static int cmd_setportbpdufilter(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if (0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if (0 > port_index)
        return port_index;
    return set_port_cfg(bpdu_filter_port, getyesno(argv[3], "yes", "no"));
}

static int cmd_setportautoedge(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if(0 > port_index)
        return port_index;
    return set_port_cfg(auto_edge_port, getyesno(argv[3], "yes", "no"));
}

static int cmd_setportp2p(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if(0 > port_index)
        return port_index;
    const char *opts[] = { "no", "yes", "auto", NULL };
    int vals[] = { p2pForceFalse, p2pForceTrue, p2pAuto };
    return set_port_cfg(admin_p2p, vals[getenum(argv[3], opts)]);
}

static int cmd_setportrestrrole(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if(0 > port_index)
        return port_index;
    return set_port_cfg(restricted_role, getyesno(argv[3], "yes", "no"));
}

static int cmd_setportrestrtcn(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if(0 > port_index)
        return port_index;
    return set_port_cfg(restricted_tcn, getyesno(argv[3], "yes", "no"));
}

static int cmd_settreeportprio(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if(0 > port_index)
        return port_index;
    int mstid = get_id(argv[3], "mstid", MAX_MSTID);
    if(0 > mstid)
        return mstid;
    unsigned int prio = getuint(argv[4]);
    if(prio > 255)
        prio = 255;
    return set_tree_port_cfg(port_priority, prio);
}

static int cmd_settreeportcost(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if(0 > port_index)
        return port_index;
    int mstid = get_id(argv[3], "mstid", MAX_MSTID);
    if(0 > mstid)
        return mstid;
    unsigned int path_cost = getuint(argv[4]);
    if(path_cost > 210000000)
        path_cost = 210000000;
    return set_tree_port_cfg(admin_internal_port_path_cost, path_cost);
}

static int cmd_portmcheck(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int port_index = get_index(argv[2], "port");
    if(0 > port_index)
        return port_index;
    return CTL_port_mcheck(br_index, port_index);
}

static int cmd_debuglevel(int argc, char *const *argv)
{
    unsigned int level = getuint(argv[1]);
    if (level > 4)
        level = 4;
    return CTL_set_debug_level(level);
}

static int cmd_batch_cmd(int argc, char *const *argv)
{
    unsigned int batch_cmd;

    if (!strcmp(argv[1], "start"))
        batch_cmd = 1;
    else if (!strcmp(argv[1], "stop"))
        batch_cmd = 0;
    else
        return -1;

    return CTL_set_batch_cmd(batch_cmd);
}

static int cmd_showmstilist(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    if (is_stp_userspace_bridge(argv[1]) == 0)
        return 0;
    int num_mstis = 0, i;
    __u16 mstids[MAX_IMPLEMENTATION_MSTIS + 1]; /* +1 - for the CIST */

    if(CTL_get_mstilist(br_index, &num_mstis, mstids))
        return -1;

    printf("%s list of known MSTIs:\n", argv[1]);
    for(i = 0; i < num_mstis; ++i)
        printf(" %hu", mstids[i]);
    printf("\n");

    return 0;
}

static int cmd_showmstconfid(int argc, char *const *argv)
{
    mst_configuration_identifier_t cfgid;
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    if (is_stp_userspace_bridge(argv[1]) == 0)
        return 0;
    int i;

    if(CTL_get_mstconfid(br_index, &cfgid))
        return -1;

    printf("%s MST Configuration Identifier:\n", argv[1]);
    printf("  Format Selector:      %hhu\n", cfgid.s.selector);
    printf("  Configuration Name:   %.*s\n", CONFIGURATION_NAME_LEN,
           cfgid.s.configuration_name);
    printf("  Revision Level:       %hu\n",
           __be16_to_cpu(cfgid.s.revision_level));
    printf("  Configuration Digest: ");
    for(i = 0; i < CONFIGURATION_DIGEST_LEN; ++i)
        printf("%02hhX", cfgid.s.configuration_digest[i]);
    printf("\n");

    return 0;
}

static int cmd_createtree(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int mstid = get_id(argv[2], "mstid", MAX_MSTID);
    if(0 > mstid)
        return mstid;
    return CTL_create_msti(br_index, mstid);
}

static int cmd_deletetree(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    int mstid = get_id(argv[2], "mstid", MAX_MSTID);
    if(0 > mstid)
        return mstid;
    return CTL_delete_msti(br_index, mstid);
}

static int cmd_showvid2fid(int argc, char *const *argv)
{
    __u16 vid2fid[MAX_VID + 2];
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    if (is_stp_userspace_bridge(argv[1]) == 0)
        return 0;
    if(CTL_get_vids2fids(br_index, vid2fid))
        return -1;

    printf("%s VID-to-FID allocation table:\n", argv[1]);
    int i, cur_fid;
    int interval_count;
    char first_char;
    vid2fid[MAX_VID + 1] = 0xFFFF; /* helps to finalize last interval */
    do{
        cur_fid = vid2fid[1];
        for(i = 1; i <= MAX_VID; ++i)
            if(cur_fid > vid2fid[i])
                cur_fid = vid2fid[i];
        if(cur_fid > MAX_FID)
            break;
        printf("  FID %u:", cur_fid);
        first_char = ' ';
        for(i = 1, interval_count = 0; i <= (MAX_VID + 1); ++i)
        {
            if(cur_fid != vid2fid[i])
            {
                if(interval_count)
                {
                    printf("%c%u", first_char, i - interval_count);
                    first_char = ',';
                    if(1 < interval_count)
                        printf("-%u", i - 1);
                    interval_count = 0;
                }
                continue;
            }
            vid2fid[i] = 0xFFFF;
            ++interval_count;
        }
        printf("\n");
    }while(true);

    return 0;
}

static int cmd_showfid2mstid(int argc, char *const *argv)
{
    __u16 fid2mstid[MAX_FID + 2];
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    if (is_stp_userspace_bridge(argv[1]) == 0)
        return 0;
    if(CTL_get_fids2mstids(br_index, fid2mstid))
        return -1;

    printf("%s FID-to-MSTID allocation table:\n", argv[1]);
    int i, cur_mstid;
    int interval_count;
    char first_char;
    fid2mstid[MAX_FID + 1] = 0xFFFF; /* helps to finalize last interval */
    do{
        cur_mstid = fid2mstid[0];
        for(i = 1; i <= MAX_FID; ++i)
            if(cur_mstid > fid2mstid[i])
                cur_mstid = fid2mstid[i];
        if(cur_mstid > MAX_MSTID)
            break;
        printf("  MSTID %u:", cur_mstid);
        first_char = ' ';
        for(i = 0, interval_count = 0; i <= (MAX_FID + 1); ++i)
        {
            if(cur_mstid != fid2mstid[i])
            {
                if(interval_count)
                {
                    printf("%c%u", first_char, i - interval_count);
                    first_char = ',';
                    if(1 < interval_count)
                        printf("-%u", i - 1);
                    interval_count = 0;
                }
                continue;
            }
            fid2mstid[i] = 0xFFFF;
            ++interval_count;
        }
        printf("\n");
    }while(true);

    return 0;
}

static int ParseList(const char *str_c, __u16 *array,
                     __u16 max_index, const char *index_doc,
                     __u16 max_value, const char *value_doc,
                     bool no_zero_index)
{
    char *next, *str, *list = strdup(str_c);
    if(NULL == list)
    {
        fprintf(stderr, "Out of memory'\n");
        return -ENOMEM;
    }
    str = list;
    while(('\0' != *list) && (':' != *list))
        ++list;
    if('\0' == *list)
    {
bad_format:
        free(str);
        fprintf(stderr, "Bad format in argument: '%s'\n", str_c);
        return -1;
    }
    *list++ = '\0';
    int value = get_id(str, value_doc, max_value);
    if(0 > value)
    {
        free(str);
        return value;
    }
    int i, nn, first_index, last_index;
    bool end = false;
    next = list;
    do{
        list = next;
        while(('\0' != *next) && (',' != *next))
            ++next;
        if('\0' == *next)
            end = true;
        else
            *next++ = '\0';
        nn = 0;
        first_index = -1;
        if((1 > sscanf(list, "%d%n", &first_index, &nn)) || (first_index < 0))
        {
            /* check for '*' */
            while((' ' == *list) || ('\t' == *list))
                ++list;
            if('*' != *list)
                goto bad_format;
            for(i = (no_zero_index ? 1 : 0); i <= max_index; ++i)
                if(array[i] > max_value)
                    array[i] = value;
            continue;
        }
        if((first_index > max_index) || (no_zero_index && (0 == first_index)))
        {
bad_index:
            free(str);
            fprintf(stderr, "Bad %s %d in argument: '%s'\n",
                    index_doc, first_index, str_c);
            return -1;
        }
        list += nn;
        while(('\0' != *list) && (('0' > *list) || ('9' < *list)))
            ++list;
        if('\0' == *list)
            last_index = first_index;
        else
        {
            last_index = -1;
            if((1 > sscanf(list, "%d", &last_index)) || (last_index < 0))
                goto bad_format;
            if((last_index > max_index) || (no_zero_index && (0 == last_index)))
            {
                first_index = last_index; /* for proper error string */
                goto bad_index;
            }
        }
        if(first_index > last_index)
        {
            i = first_index;
            first_index = last_index;
            last_index = i;
        }
        for(i = first_index; i <= last_index; ++i)
            array[i] = value;
    }while(!end);

    free(str);
    return 0;
}

static int cmd_setvid2fid(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    __u16 vids2fids[MAX_VID + 1];
    memset(vids2fids, 0xFF, sizeof(vids2fids));
    int i, ret;
    for(i = 2; i < argc; ++i)
        if(0 > (ret = ParseList(argv[i], vids2fids, MAX_VID, "VID",
                                MAX_FID, "FID", true)))
            return ret;
    return CTL_set_vids2fids(br_index, vids2fids);
}

static int cmd_setfid2mstid(int argc, char *const *argv)
{
    int br_index = get_index(argv[1], "bridge");
    if(0 > br_index)
        return br_index;
    __u16 fids2mstids[MAX_FID + 1];
    memset(fids2mstids, 0xFF, sizeof(fids2mstids));
    int i, ret;
    for(i = 2; i < argc; ++i)
        if(0 > (ret = ParseList(argv[i], fids2mstids, MAX_FID, "FID",
                                MAX_MSTID, "mstid", false)))
            return ret;
    return CTL_set_fids2mstids(br_index, fids2mstids);
}

static int cmd_showall(int argc, char *const *argv)
{
    int i, j, br_count = 0, pr_count = 0;
    struct dirent **br_namelist, **pr_namelist;
    int err = 0;
    int br_index;
    u_char use_json;
    json_object *json = NULL;

    if (argv[1] != NULL && (strcmp(argv[1], "json") == 0))
      use_json = 1;
    else
      use_json = 0;

    if (use_json)
      json = json_object_new_object();

    err = showglobal(use_json, json);

    br_count = scandir(SYSFS_CLASS_NET, &br_namelist, isbridge, sorting_func);

    if(0 > br_count)
      {
        if (use_json)
          {
            json_object_string_add(json, "error","Getting a list of all bridges");
            printf("\n%s\n", json_object_to_json_string(json));
            json_object_free(json);
          }
        else
          fprintf(stderr, "Error getting list of all bridges\n");

        return -1;
      }

    all = 1;
    detail = 1;

    for(i = 0; i < br_count; i++)
      {
        const char *br_name, *pr_name;
        json_object *json_bridge = NULL;

        if (use_json)
          json_bridge = json_object_new_object();

        br_name = br_namelist[i]->d_name;
        br_index = get_index(br_name, "bridge");

        if (0 > br_index)
          continue;

        if (use_json)
          json_object_int_add(json_bridge, "bridgeIndex", br_index);
        else
          printf("\nBRIDGE: %s, Br_index: %d\n", br_name, br_index);

        err = do_showbridge(br_name, PARAM_NULL, use_json, json_bridge);

        if (err)
          continue;

        pr_count = get_port_list(br_name, &pr_namelist);

        for (j = 0; j < pr_count; j++)
          {
            json_object *json_port = NULL;
            pr_name = pr_namelist[j]->d_name;

            if (use_json)
              json_port = json_object_new_object();

            err = do_showport(br_index, br_name, pr_name, PARAM_NULL, use_json, json_port);

            if (use_json)
              json_object_object_add(json_bridge, pr_name, json_port);

            if (err)
              continue;
          }

        for (j = 0; j < pr_count; j++)
          free(pr_namelist[j]);

        free(pr_namelist);

        if (use_json)
          json_object_object_add(json, br_name, json_bridge);
      }

    if (!use_json)
      printf("\n");

    for (i = 0; i < br_count; i++)
      free(br_namelist[i]);

    free(br_namelist);
    all = 0;
    detail = 0;

    if (use_json)
      {
        printf("\n%s\n", json_object_to_json_string(json));
        json_object_free(json);
      }

    return err;
}


struct command
{
    int nargs;
    int optargs;
    const char *name;
    int (*func) (int argc, char *const *argv);
    const char *format;
    const char *help;
};

static const struct command commands[] =
{
    /* Add/delete bridges */
    {1, 32, "addbridge", cmd_addbridge,
     "<bridge> [<bridge> ...]", "Add bridges to the mstpd's list"},
    {1, 32, "delbridge", cmd_delbridge,
     "<bridge> [<bridge> ...]", "Remove bridges from the mstpd's list"},

    /* Show global bridge */
    {0, 33, "showbridge", cmd_showbridge,
     "[json] [<bridge> ... [param]]", "Show bridge state for the CIST"},
    {1, 0, "showmstilist", cmd_showmstilist,
     "<bridge>", "Show list of registered MSTIs"},
    {1, 0, "showmstconfid", cmd_showmstconfid,
     "<bridge>", "Show MST ConfigId"},
    {1, 0, "showvid2fid", cmd_showvid2fid,
     "<bridge>", "Show VID-to-FID allocation table"},
    {1, 0, "showfid2mstid", cmd_showfid2mstid,
     "<bridge>", "Show FID-to-MSTID allocation table"},
    /* Show global port */
    {1, 33, "showport", cmd_showport,
     "<bridge> [json] [<port> ... [param]]", "Show port state for the CIST"},
    {1, 33, "showportdetail", cmd_showportdetail,
     "<bridge> [json] [<port> ... [param]]", "Show port detailed state for the CIST"},
    /* Show tree bridge */
    {2, 0, "showtree", cmd_showtree,
     "<bridge> <mstid>", "Show bridge state for the given MSTI"},
    /* Show tree port */
    {3, 0, "showtreeport", cmd_showtreeport,
     "<bridge> <port> <mstid>", "Show port detailed state for the given MSTI"},

    /* Set global bridge */
    {3, 0, "setmstconfid", cmd_setmstconfid,
     "<bridge> <revision> <name>",
     "Set MST ConfigId elements: Revision Level (0-65535) and Name"},
    {2, 32, "setvid2fid", cmd_setvid2fid,
     "<bridge> <FID>:<VIDs List>",
     "Set VIDs-to-FIDs allocation"},
    {2, 32, "setfid2mstid", cmd_setfid2mstid,
     "<bridge> <mstid>:<FIDs List>",
     "Set FIDs-to-MSTIDs allocation"},
    {2, 0, "setmaxage", cmd_setbridgemaxage,
     "<bridge> <max_age>", "Set bridge max age (6-40)"},
    {2, 0, "setfdelay", cmd_setbridgefdelay,
     "<bridge> <fwd_delay>", "Set bridge forward delay (4-30)"},
    {2, 0, "setmaxhops", cmd_setbridgemaxhops,
     "<bridge> <max_hops>", "Set bridge max hops (6-40)"},
    {2, 0, "setforcevers", cmd_setbridgeforcevers,
     "<bridge> {mstp|rstp|stp}", "Force Spanning Tree protocol version"},
    {2, 0, "settxholdcount", cmd_setbridgetxholdcount,
     "<bridge> <tx_hold_count>", "Set bridge transmit hold count (1-10)"},
    /* Set tree bridge */
    {2, 0, "createtree", cmd_createtree,
     "<bridge> <mstid>", "Create new MSTI"},
    {2, 0, "deletetree", cmd_deletetree,
     "<bridge> <mstid>", "Delete existing MSTI"},
    {3, 0, "settreeprio", cmd_settreeprio,
     "<bridge> <mstid> <priority>",
     "Set bridge priority (0-65535) for the given MSTI"},
    /* Set global port */
    {3, 0, "setportpathcost", cmd_setportpathcost,
     "<bridge> <port> <cost>",
     "Set port external path cost for the CIST (0 = auto)"},
    {3, 0, "setportadminedge", cmd_setportadminedge,
     "<bridge> <port> {yes|no}", "Set initial edge state"},
    {3, 0, "setportautoedge", cmd_setportautoedge,
     "<bridge> <port> {yes|no}", "Enable auto transition to/from edge state"},
    {3, 0, "setportp2p", cmd_setportp2p,
     "<bridge> <port> {yes|no|auto}", "Set p2p detection mode"},
    {3, 0, "setportrestrrole", cmd_setportrestrrole,
     "<bridge> <port> {yes|no}", "Restrict port ability to take Root role"},
    {3, 0, "setportrestrtcn", cmd_setportrestrtcn,
     "<bridge> <port> {yes|no}",
     "Restrict port ability to propagate received TCNs"},
    {2, 0, "portmcheck", cmd_portmcheck,
     "<bridge> <port>", "Try to get back from STP to rapid (RSTP/MSTP) mode"},
    {3, 0, "setbpduguard", cmd_setportbpduguard,
     "<bridge> <port> {yes|no}", "Set bpdu guard state"},
    /* Set tree port */
    {4, 0, "settreeportprio", cmd_settreeportprio,
     "<bridge> <port> <mstid> <priority>",
     "Set port priority (0-240) for the given MSTI"},
    {4, 0, "settreeportcost", cmd_settreeportcost,
     "<bridge> <port> <mstid> <cost>",
     "Set port internal path cost for the given MSTI (0 = auto)"},
    {0, 1, "showall", cmd_showall,
     "[json]", "Show detailed mstpctl info"},
    {2, 0, "sethello", cmd_setbridgehello,
     "<bridge> <hello_time>", "Set bridge hello time (1-10)"},
    {2, 0, "setageing", cmd_setbridgeageing,
     "<bridge> <ageing_time>", "Set bridge ageing time (10-1000000)"},
    {3, 0, "setportnetwork", cmd_setportnetwork,
     "<bridge> <port> {yes|no}", "Set port network state"},
    {3, 0, "setportdonttxmt", cmd_setportdonttxmt,
     "<bridge> <port> {yes|no}", "Disable/Enable sending BPDU"},
    {3, 0, "setportbpdufilter", cmd_setportbpdufilter,
     "<bridge> <port> {yes|no}", "Set BPDU filter state"},

    /* Other */
    {1, 0, "debuglevel", cmd_debuglevel, "<level>", "Level of verbosity (1-4)"},
    {1, 0, "batch", cmd_batch_cmd, "{start/stop}", "Start/Stop batching of commands"},
};

static const struct command *command_lookup(const char *cmd)
{
    int i;

    for(i = 0; i < COUNT_OF(commands); ++i)
    {
        if(!strcmp(cmd, commands[i].name))
            return &commands[i];
    }

    return NULL;
}

static void command_helpall(void)
{
    int i;

    for(i = 0; i < COUNT_OF(commands); ++i)
    {
        if((strcmp("setportdonttxmt", commands[i].name)) &&
                (strcmp("batch", commands[i].name)))
            printf("\t%-22s%-36s %s\n", commands[i].name,
                commands[i].format, commands[i].help);
    }
}

static void help(void)
{
    printf("Usage: mstpctl [commands]\n");
    printf("commands:\n");
    command_helpall();
}

#define PACKAGE_VERSION2(v, b) "mstp, " #v "-" #b
#define PACKAGE_VERSION(v, b) PACKAGE_VERSION2(v, b)

int main(int argc, char *const *argv)
{
    const struct command *cmd;
    int f;
    static const struct option options[] =
    {
        {.name = "help",    .val = 'h'},
        {.name = "version", .val = 'V'},
        {0}
    };

    while(EOF != (f = getopt_long(argc, argv, "Vh", options, NULL)))
        switch(f)
        {
            case 'h':
                help();
                return 0;
            case 'V':
                printf("%s\n", PACKAGE_VERSION(VERSION, BUILD));
                return 0;
            default:
                fprintf(stderr, "Unknown option '%c'\n", f);
                goto help;
        }

    if(argc == optind)
        goto help;

    if(ctl_client_init())
    {
        fprintf(stderr, "can't setup control connection\n");
        return 1;
    }

    argc -= optind;
    argv += optind;
    if(NULL == (cmd = command_lookup(argv[0])))
    {
        fprintf(stderr, "never heard of command [%s]\n", argv[0]);
        goto help;
    }

    if(argc < cmd->nargs + 1 || argc > cmd->nargs + cmd->optargs + 1)
    {
        printf("Incorrect number of arguments for command\n");
        printf("Usage: mstpctl %s %s\n  %s\n",
               cmd->name, cmd->format, cmd->help);
        return 1;
    }

    return cmd->func(argc, argv);

help:
    help();
    return 1;
}

/* Implementation of client-side functions */
CLIENT_SIDE_FUNCTION(get_global_status)
CLIENT_SIDE_FUNCTION(get_cist_bridge_status)
CLIENT_SIDE_FUNCTION(get_msti_bridge_status)
CLIENT_SIDE_FUNCTION(set_cist_bridge_config)
CLIENT_SIDE_FUNCTION(set_msti_bridge_config)
CLIENT_SIDE_FUNCTION(get_cist_port_status)
CLIENT_SIDE_FUNCTION(get_msti_port_status)
CLIENT_SIDE_FUNCTION(set_cist_port_config)
CLIENT_SIDE_FUNCTION(set_msti_port_config)
CLIENT_SIDE_FUNCTION(port_mcheck)
CLIENT_SIDE_FUNCTION(set_debug_level)
CLIENT_SIDE_FUNCTION(set_clag_role)
CLIENT_SIDE_FUNCTION(set_clag_system_mac)
CLIENT_SIDE_FUNCTION(get_mstilist)
CLIENT_SIDE_FUNCTION(create_msti)
CLIENT_SIDE_FUNCTION(delete_msti)
CLIENT_SIDE_FUNCTION(get_mstconfid)
CLIENT_SIDE_FUNCTION(set_mstconfid)
CLIENT_SIDE_FUNCTION(get_vids2fids)
CLIENT_SIDE_FUNCTION(get_fids2mstids)
CLIENT_SIDE_FUNCTION(set_vid2fid)
CLIENT_SIDE_FUNCTION(set_fid2mstid)
CLIENT_SIDE_FUNCTION(set_vids2fids)
CLIENT_SIDE_FUNCTION(set_fids2mstids)
CLIENT_SIDE_FUNCTION(set_batch_cmd)

CTL_DECLARE(add_bridges)
{
    int res = 0;
    LogString log = { .buf = "" };
    int i, chunk_count, brcount, serialized_data_count;
    int *serialized_data, *ptr;

    chunk_count = serialized_data_count = (brcount = br_array[0]) + 1;
    for(i = 0; i < brcount; ++i)
        serialized_data_count += ifaces_lists[i][0] + 1;
    if(NULL == (serialized_data = malloc(serialized_data_count * sizeof(int))))
    {
        LOG("out of memory, serialized_data_count = %d",
            serialized_data_count);
        return -1;
    }
    memcpy(serialized_data, br_array, chunk_count * sizeof(int));
    ptr = serialized_data + chunk_count;
    for(i = 0; i < brcount; ++i)
    {
        chunk_count = ifaces_lists[i][0] + 1;
        memcpy(ptr, ifaces_lists[i], chunk_count * sizeof(int));
        ptr += chunk_count;
    }

    int r = send_ctl_message(CMD_CODE_add_bridges, serialized_data,
                             serialized_data_count * sizeof(int),
                             NULL, 0, &log, &res);
    free(serialized_data);
    if(r || res)
        LOG("Got return code %d, %d\n%s", r, res, log.buf);
    if(r)
        return r;
    if(res)
        return res;
    return 0;
}

CTL_DECLARE(del_bridges)
{
    int res = 0;
    LogString log = { .buf = "" };
    int r = send_ctl_message(CMD_CODE_del_bridges,
                             br_array, (br_array[0] + 1) * sizeof(int),
                             NULL, 0, &log, &res);
    if(r || res)
        LOG("Got return code %d, %d\n%s", r, res, log.buf);
    if(r)
        return r;
    if(res)
        return res;
    return 0;
}


/*********************** Logging *********************/

void Dprintf(int level, const char *fmt, ...)
{
    char logbuf[LOG_STRING_LEN];
    logbuf[sizeof(logbuf) - 1] = 0;
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(logbuf, sizeof(logbuf) - 1, fmt, ap);
    va_end(ap);
    printf("%s\n", logbuf);
}
