/*****************************************************************************
  Copyright (c) 2006 EMC Corporation.
  Copyright (c) 2011 Factor-SPE

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  The full GNU General Public License is included in this distribution in the
  file called LICENSE.

  Authors: Srinivas Aji <Aji_Srinivas@emc.com>
  Authors: Vitalii Demianets <vitas@nppfactor.kiev.ua>

******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdbool.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <linux/if_ether.h>
#include <linux/ethtool.h>
#include <linux/sockios.h>
#include <errno.h>

#include "log.h"

static int netsock = -1;

int netsock_init(void)
{
    netsock = socket(AF_INET, SOCK_DGRAM, 0);
    if(0 > netsock)
    {
        ERROR("Couldn't open inet socket for ioctls: %m\n");
        return -1;
    }
    return 0;
}

int get_hwaddr(char *ifname, __u8 *hwaddr)
{
    struct ifreq ifr;
    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    if(0 > ioctl(netsock, SIOCGIFHWADDR, &ifr))
    {
        INFO("%s: get hw address failed: %m", ifname);
        return -1;
    }
    memcpy(hwaddr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);
    return 0;
}

int get_flags(char *ifname)
{
    struct ifreq ifr;

    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    if(0 > ioctl(netsock, SIOCGIFFLAGS, &ifr))
    {
        INFO("%s: get interface flags failed: %m", ifname);
        return -1;
    }
    return ifr.ifr_flags;
}

int set_if_down(char *ifname)
{
    struct ifreq ifr;
    int br_flags;

    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    br_flags = get_flags(ifname);
    if (br_flags < 0) {
        return -1;
    }
    ifr.ifr_flags = br_flags & ~IFF_UP;
    if (0 > ioctl(netsock, SIOCSIFFLAGS, &ifr))
    {
        ERROR("%s: set if down flag failed: %m", ifname);
        return -1;
    }
    return 0;
}

int ethtool_get_speed_duplex(char *ifname, int *speed, int *duplex)
{
    struct ifreq ifr;

    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    struct ethtool_cmd ecmd;

    ecmd.cmd = ETHTOOL_GSET;
    ifr.ifr_data = (caddr_t)&ecmd;
    if(0 > ioctl(netsock, SIOCETHTOOL, &ifr))
    {
        /* for vxlan device since duplex/speed is not set,
         * hardcode to full duplex/10G for now, eopnotsupp
         * error code can be used as an indication for this case.
         */
        if (errno == EOPNOTSUPP) {
            LOG("unsupported: get speed/duplex for %s: %m\n", ifname);
            *speed = 10000;
            *duplex = 1;
            return 0;
        }
        INFO("Cannot get speed/duplex for %s: %m\n", ifname);
        return -1;
    }
    *speed = (ecmd.speed_hi << 16)  | ecmd.speed;   /* Ethtool speed is in Mbps */
    *duplex = ecmd.duplex; /* We have same convention as ethtool.
                               0 = half, 1 = full */
    return 0;
}

/********* Sysfs based utility functions *************/

/* This sysfs stuff might break with interface renames */
bool is_bridge(char *if_name)
{
    char path[32 + IFNAMSIZ];
    sprintf(path, "/sys/class/net/%s/bridge", if_name);
    return (0 == access(path, R_OK));
}

int get_bridge_portno(char *if_name)
{
    char path[32 + IFNAMSIZ];
    sprintf(path, "/sys/class/net/%s/brport/port_no", if_name);
    char buf[128];
    int fd;
    long res = -1;
    TSTM((fd = open(path, O_RDONLY)) >= 0, -1, "%m");
    int l;
    TSTM((l = read(fd, buf, sizeof(buf) - 1)) >= 0, -1, "%m");
    if(0 == l)
    {
        ERROR("Empty port index file");
        goto out;
    }
    else if((sizeof(buf) - 1) == l)
    {
        ERROR("port_index file too long");
        goto out;
    }
    buf[l] = 0;
    if('\n' == buf[l - 1])
        buf[l - 1] = 0;
    char *end;
    res = strtoul(buf, &end, 0);
    if(0 != *end || INT_MAX < res)
    {
        ERROR("Invalid port index %s", buf);
        res = -1;
    }
out:
    close(fd);
    return res;
}

/* Find if interface is a vlan device and if so,
 * find the vlan id from /proc/net/vlan/ifname
 */
bool is_vlan(char *if_name, int *vlan, char *vlan_bond)
{
    char path[32 + IFNAMSIZ];
    int err;
    int rc;
    FILE *fp = NULL;
    char dev[IFNAMSIZ];
    char vlan_str[IFNAMSIZ];
    int vlan_id = 0;
    char *line = NULL;
    size_t len;
    int i;
    ssize_t read;

    *vlan = 0;
    sprintf(path, "/proc/net/vlan/%s", if_name);
    err = access(path, F_OK);
    if (err != 0) {
        goto out;
    }
	fp = fopen(path, "r");
	if (fp == NULL) {
        ERROR("Cannot open file %s %d %s", path, errno, strerror(errno));
        goto out;
    }
	rc = fscanf(fp, "%s %s %d\n", dev, vlan_str, &vlan_id);
	if ((rc != 3) && (rc != EOF)) {
        ERROR("got %d arg", rc);
		goto out;
    }
    *vlan = vlan_id;
    if (!vlan_bond)
        goto out;

    for (i = 0; i < 8; i++) {
        read = getline(&line, &len, fp);
        if (read == -1)
            goto out;
    }
    sscanf(line, "%s %s", dev, vlan_bond);
out:
    if (fp)
        fclose(fp);
    if (line)
        free(line);
    return (err == 0);
}

/* find if vlan filtering is enabled for bridge
 */
bool is_bridge_vlan_filter(char *br_name)
{
    char path[52 + IFNAMSIZ];
    int err;
    FILE *fp = NULL;
    int filter;

    sprintf(path, "/sys/class/net/%s/bridge/vlan_filtering", br_name);
    err = access(path, F_OK);
    if (err != 0)
        goto out;
    fp = fopen(path, "r");
    if (fp == NULL)
    {
        ERROR("Cannot open file %s %d %s", path, errno, strerror(errno));
        goto out;
    }
    err = fscanf(fp, "%d\n", &filter);
    if ((err != 1) && (err != EOF))
    {
        ERROR("got %d arg", err);
        goto out;
    }
out:
    if (fp)
        fclose(fp);
    return filter;
}
